/***************************************************************************/
/* 			     		    	                           */
/*  TWO DIMENSIONAL GENERALIZED Edge DETECTOR           		   */
/* 			     		    	                           */
/*  B. Kurt, M. Gokmen, "Two Dimensional Generalized Edge Detector,"       */
/*          appear to the 10th International Conference on Image Analysis  */
/*                           and Processing, ICIAP'99, Venice Italy, 1999. */
/* 			     		    	                           */
/***************************************************************************/
/* 			     		    	                           */
/*  Edge Class Library       		                                   */
/* 			     		    	                           */
/*  Edge              	     		    	                           */
/*    |              	     		    	                           */
/*  image              	     		    	                           */
/* 			     		    	                           */
/***************************************************************************/
/* 			     		    	                           */
/*  Last modified by    		    	                           */
/*                   Binnur KURT            	                           */
/*  When        	     		    	                           */
/* 		     August 19th 1999    	                           */
/* 			     		    	                           */
/***************************************************************************/
#include <math>

#include "Edge.h"
#include "Image.h"

Edge::Edge(){
       AreEdgesDetected = false                         ;
}

     Edge::~Edge(){
     FreeTemporary()                                    ;
}

void Edge::FreeTemporary(){
     int i ;

     if (AreEdgesDetected==true){
          delete[] Dx[0] ;
          delete[] Dy[0] ;
          delete[] Mag[0] ;
          delete[] temp[0] ;
          delete[] Dx;
          delete[] Dy;
          delete[] Mag;
          delete[] temp;
     }
     AreEdgesDetected= false ;
}

void Edge::SetGFilterLength(int filterlength){
     FilterLength = filterlength ;
}

void Edge::SaveEdgesToTImage(TImage *img){
     static int i,j;
     static unsigned char *p,*q,cl ;
     Graphics::TBitmap *myBitmap =new Graphics::TBitmap() ;
     myBitmap->Width=Column ;
     myBitmap->Height=Row ;
     myBitmap->PixelFormat = pf24bit  ;

     img->Hint = "Edge Map"                                    ;

     p= edge[0];
     for (i=0;i<Row;i++){
         q= (unsigned char*) myBitmap->ScanLine[i];
         for (j=0;j<Column;j++,p++){
             cl = *p ? 255:0 ;
             *q++ = cl ;
             *q++ = cl ;
             *q++ = cl ;
         }
     }
     img->Picture->Bitmap->Assign(myBitmap);
     delete myBitmap;
     img->Refresh();
}

void Edge::SetThresholds(int low,int high){
     TLow = low ; THigh = high ;
}

void Edge::SetLambda(float lambda){
     Lambda = lambda ;
}

void Edge::SetTau(float tau){
     Tau = tau ;
}

void Edge::GetRFilter(){
    int i,j,k,l;
    float abs_ij,a,b,kx,K,Kx,Ky,psi,theta;
    float A,B,Delta;
    float *p;

    filter.Gain = 1.0                   ;
    filter.xx = FilterLength            ;
    filter.yy = FilterLength            ;
    filter.Width  = 2*FilterLength+1    ;
    filter.Height = 2*FilterLength+1    ;


    filter.coeff = new float*[filter.Height] ;
    filter.coeff[0] = new float[filter.Height*filter.Width] ;
    p=filter.coeff[0]+filter.Width;
    for (i=1;i<filter.Height;i++,p+=filter.Width)
        filter.coeff[i] = p ;

    A = Lambda * Tau                    ;
    B = Lambda * (1-Tau)                ;
    Delta = B * B - 4.0 * A             ;

    if(A==0.0){
    K = 0.5/B                                 ;
    kx = - 1.0 / sqrt(2.0*B)    ;
    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] = K * exp(kx*abs_ij)        ;
       }
    }else if(Delta>0.0){
    a = sqrt( ( B + sqrt(Delta) ) / ( 4.0 * A )) ;
    b = sqrt( ( B - sqrt(Delta) ) / ( 4.0 * A )) ;
    K = (2.0 * A) / sqrt(Delta) 				;
    Kx = - K * b                                ;
    Ky = K * a                                  ;
    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] = Kx * exp(-a*abs_ij) +
                              Ky * exp(-b*abs_ij) ;
       }
    }else if(Delta<0.0){
     if(B == 0.0)
       theta = M_PI / 4.0 ;
     else
       theta = 0.5 * atan(sqrt(fabs(Delta))/B)                                  ;
     K     = 4.0 / (sqrt(A)*cos(theta)*cos(theta)*cos(theta)*
                            sin(2.0*theta)*sin(2.0*theta))              ;
     psi   = sin(theta) / (sqrt(sqrt(4.0*A)))                                   ;
     Kx    = K * cos(theta)                                                                     ;
     Ky    = K * sin(theta)                                                                     ;
     kx    = - cos(theta) / (sqrt(sqrt(4.0*A)))                 ;

    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] = ( Kx * cos(psi*abs_ij) +
                                Ky * sin(psi*abs_ij) ) * exp(kx*abs_ij);
       }
    }
    else if(Delta==0.0){
    a = 1.0  / ( sqrt(B) )                     ;
    kx = a                                                         ;
    K  = 1.0 / (4.0 * B)                        ;
    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] =
               K * ( kx * abs(i) * 0.5 + 1.0 ) *
                   ( kx * abs(j) * 0.5 + 1.0 )
               * exp(-kx*abs_ij) ;
       }
    }
    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++)
          filter.Gain += filter.coeff[i+FilterLength][j+FilterLength] ;
    filter.Gain = 1.0 / filter.Gain ;
}

void Edge::GetGxFilter(){
    int i,j,k,l                                                         ;
    float abs_ij,a,b,kx,K,Kx,Ky,psi,theta                               ;
    float A,B,Delta                                                     ;
    float *p;
    
    filter.Gain= 1.0                   ;
    filter.xx= FilterLength            ;
    filter.yy= FilterLength            ;
    filter.Width= 2*FilterLength+1    ;
    filter.Height= 2*FilterLength+1    ;


    filter.coeff = new float*[filter.Height] ;
    filter.coeff[0] = new float[filter.Height*filter.Width] ;
    p=filter.coeff[0]+filter.Width;
    for (i=1;i<filter.Height;i++,p+=filter.Width)
        filter.coeff[i] = p ;


    B = Lambda * (1-Tau)                ;
    A = Lambda * Tau                    ;
    Delta = B * B - 4.0 * A             ;

    if(A==0.0){
    K = -0.5/ (B * sqrt(2.0*B))                 ;
    kx = - 1.0 / sqrt(2.0*B)    		;
    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] = K * Signum(i) * exp(kx*abs_ij)        ;
       }
    }else if(Delta>0.0){
    a = sqrt( ( B + sqrt(Delta) ) / ( 4.0 * A )) ;
    b = sqrt( ( B - sqrt(Delta) ) / ( 4.0 * A )) ;
    K = (2.0 * A) / sqrt(Delta) 		 ;
    Kx = K * b * a                               ;
    Ky = - K * a * b                             ;
    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] = Signum(i) * (Kx * exp(-a*abs_ij) +
                              Ky * exp(-b*abs_ij)) ;
       }
    }else if(Delta<0.0){
     if(B == 0.0)
       theta = M_PI / 4.0 ;
     else
       theta = 0.5 * atan(sqrt(fabs(Delta))/B)                                  ;
     K     = 4.0 / (sqrt(A)*cos(theta)*cos(theta)*cos(theta)*
                            sin(2.0*theta)*sin(2.0*theta))              ;
     psi   = sin(theta) / (sqrt(sqrt(4.0*A)))                                   ;
     Kx    = K * (sin(theta) * psi - (1.0 / sqrt(2.0*sqrt(A))) * cos(theta))              ;
     Ky    = K * (sin(theta) * (1.0 / sqrt(2.0*sqrt(A))) + psi * cos(theta) )    ;
     kx    = - cos(theta) / (sqrt(sqrt(4.0*A)))                 ;

    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] = Signum(i) * ( Kx * cos(psi*abs_ij) +
                                Ky * sin(psi*abs_ij) ) * exp(kx*abs_ij);
       }
    }
    else if(Delta==0.0){
    a = 1.0  / ( sqrt(B) )                     ;
    kx = a                                                         ;
    K  = -1.0 / (4.0 * B * B)                        ;
    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] =
               K * Signum(i) * (float) k *
                   ( kx * abs(j) + 1.0 )
               * exp(-kx*abs_ij) ;
       }
    }
}

void Edge::GetGyFilter(){
    int i,j,k,l;
    float abs_ij,a,b,kx,K,Kx,Ky,psi,theta;
    float A,B,Delta,*p;

    filter.Width= 2*FilterLength+1;
    filter.Height= 2*FilterLength+1;
    filter.Gain= 1.0;
    filter.xx= FilterLength;
    filter.yy= FilterLength;
    // memory allocation
    filter.coeff= new float*[filter.Height];
    filter.coeff[0]= new float[filter.Height*filter.Height];
    p= filter.coeff[0]+filter.Width ;
    for (i=1;i<filter.Height;i++,p+=filter.Width)
        filter.coeff[i] = p ;

    B = Lambda * (1-Tau)                ;
    A = Lambda * Tau                    ;
    Delta = B * B - 4.0 * A             ;

    if(A==0.0){
    K = -0.5/ (B * sqrt(2.0*B))                 ;
    kx = - 1.0 / sqrt(2.0*B)    ;
    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] = K * Signum(j) * exp(kx*abs_ij)        ;
       }
    }else if(Delta>0.0){
    a = sqrt( ( B + sqrt(Delta) ) / ( 4.0 * A )) ;
    b = sqrt( ( B - sqrt(Delta) ) / ( 4.0 * A )) ;
    K = (2.0 * A) / sqrt(Delta) 		 ;
    Kx = K * b * a                               ;
    Ky = - K * a * b                             ;
    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] = Signum(j) * (Kx * exp(-a*abs_ij) +
                              Ky * exp(-b*abs_ij)) ;
       }
    }else if(Delta<0.0){
     if(B == 0.0)
       theta = M_PI / 4.0 ;
     else
       theta = 0.5 * atan(sqrt(fabs(Delta))/B)                                  ;
     K     = 4.0 / (sqrt(A)*cos(theta)*cos(theta)*cos(theta)*
                            sin(2.0*theta)*sin(2.0*theta))              ;
     psi   = sin(theta) / (sqrt(sqrt(4.0*A)))                                   ;
     Kx    = K * (sin(theta) * psi - (1.0 / sqrt(2.0*sqrt(A))) * cos(theta))              ;
     Ky    = K * (sin(theta) * (1.0 / sqrt(2.0*sqrt(A))) + psi * cos(theta) )    ;
     kx    = - cos(theta) / (sqrt(sqrt(4.0*A)))                 ;

    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] = Signum(j) * ( Kx * cos(psi*abs_ij) +
                                Ky * sin(psi*abs_ij) ) * exp(kx*abs_ij);
       }
    }
    else if(Delta==0.0){
    a = 1.0  / ( sqrt(B) )                     ;
    kx = a                                                         ;
    K  = -1.0 / (4.0 * B * B)                        ;
    for(i=-FilterLength;i<=FilterLength;i++)
       for(j=-FilterLength;j<=FilterLength;j++){
          k = i<0 ? -i:i ;
          l = j<0 ? -j:j ;
          abs_ij = (float)  k+ (float) l  ;
          filter.coeff[i+FilterLength][j+FilterLength] =
               K * Signum(j) * (float) l *
                   ( kx * abs(i) + 1.0 )
               * exp(-kx*abs_ij) ;
       }
    }
}

inline float Edge::Signum(int v)
{
      if(v>0) return 1.0 ;
      else if(v<0) return -1.0 ;
      else return(0.0)	;
}

void Edge::RFilter(){
     int i,j;
     float *p ;
     unsigned char *q;
     GetRFilter();
     Convolve();

     p= filterOutput[0];

     q= Y[0];

     for(i=0;i<Row;i++)
        for (j=0;j<Column;j++,p++,q++)

            if (*p>255.0) *q = 255 ;

            else

            if (*p<0.0) *q= 0 ;

            else *q= *p ;

}

void Edge::GetEdges(){
    int i,j     ;

    if(IsNewImageLoaded == true)
      FreeTemporary()   ;

    if (AreEdgesDetected==false){
       float *p ;
       unsigned char *q;
       Dx = new float *[Row] ;
       Dx[0] = new float [Row*Column] ;
       p= Dx[0] + Column ;
       for (i=1;i<Row;i++,p+=Column)
           Dx[i] = p ;

       Dy = new float *[Row] ;
       Dy[0] = new float [Row*Column] ;
       p= Dy[0] + Column ;
       for (i=1;i<Row;i++,p+=Column)
           Dy[i] = p ;

       Mag = new float *[Row] ;
       Mag[0] = new float [Row*Column] ;
       p= Mag[0] + Column ;
       for (i=1;i<Row;i++,p+=Column)
           Mag[i] = p ;
       edge = new unsigned char *[Row] ;
       edge[0] = new unsigned char [Row*Column] ;
       q= edge[0] + Column ;
       for (i=1;i<Row;i++,q+=Column)
           edge[i] = q ;
       temp = new unsigned char *[Row] ;

       temp[0] = new unsigned char [Row*Column] ;
       q= temp[0] + Column ;
       for (i=1;i<Row;i++,q+=Column)
           temp[i] = q ;
    }

    AreEdgesDetected= true                             ;

    /* Convolution Along x-direction */

    GetGxFilter();

    Convolve();

    memcpy(Dy[0],filterOutput[0],sizeof(float)*Row*Column);

    /* Convolution Along y-direction */

    GetGyFilter();

    Convolve();

    memcpy(Dx[0],filterOutput[0],sizeof(float)*Row*Column);

    /* non-maximum suppression */

    NonMaxSuppression() ;

    TwoLevelThresholding() ;

    ClearBorders();

    /* thinning */

    thinning() ;

}


void Edge::NonMaxSuppression(){

      int i,j;

      float mag1,mag2,z1,z2,xperp,yperp,m00;
      float *pMag,*pDx,*pDy;

      for (i=0;i<Row;i++){
          Dx[i][0]= 0.0 ;
          Dy[i][0]= 0.0 ;
          Dx[i][Column-1] = 0.0 ;
          Dy[i][Column-1] = 0.0 ;
      }
      memset(Dx[0],0,sizeof(float)*Column) ;
      memset(Dy[0],0,sizeof(float)*Column) ;
      memset(Dx[Row-1],0,sizeof(float)*Column) ;
      memset(Dy[Row-1],0,sizeof(float)*Column) ;

      pMag= Mag[0];
      pDx= Dx[0];
      pDy= Dy[0];
      register float v1,v2 ;
      for (i=0;i<Row;i++){
          memset(temp[i],0,Column*sizeof(unsigned char)) ;
          for (j=0;j<Column;j++,pMag++,pDx++,pDy++){
              v1= *pDx;
              v2= *pDy;
              *pMag= sqrt(v1*v1+v2*v2) ;
          }
      }
      unsigned char *pTemp ;

      for (i=1;i<Row-1;i++){
          pMag= Mag[i]+1 ;
          pDx= Dx[i]+1 ;
          pDy= Dy[i]+1 ;
          pTemp= temp[i]+1;
          for (j=1;j<Column-1;j++,pMag++,pTemp++,pDx++,pDy++){
              m00 = *pMag ;
              if ((int)m00==0){
                 *pTemp= NO_EDGE;
              }
              else {
                xperp= -*pDx/m00 ;
                yperp=  *pDy/m00 ;
              }

              if ((int)*pDx >= 0)
              {
                 if ((int)*pDy >= 0)
                 {
                    if ((int)*pDx >= (int)*pDy)
                    {
                        /* 111 */
                        /* Left point */
                        z1= *(pMag-1);
                        z2= *(pMag-Column-1);

                        mag1 = (m00 - z1)*xperp + (z2 - z1)*yperp;

                        /* Right point */
                        z1 = *(pMag+1);
                        z2= *(pMag+Column+1);

                        mag2 = (m00 - z1)*xperp + (z1 - z2)*yperp;
                    }
                    else
                    {
                        /* 110 */
                        /* Left point */
                        z1 = *(pMag-Column);
                        z2 = *(pMag-Column-1);

                        mag1 = (z1 - z2)*xperp + (z1 - m00)*yperp;

                        /* Right point */
                        z1 = *(pMag+Column);
                        z2 = *(pMag+Column+1);

                        mag2 = (z1 - z2)*xperp + (z1 - m00)*yperp;
                    }
                }
                else
                {
                    if ((int)*pDx >= -(int)*pDy)
                    {
                        /* 101 */
                        /* Left point */
                        z1 = *(pMag-1);
                        z2 = *(pMag+Column-1);

                        mag1 = (m00 - z1)*xperp + (z1 - z2)*yperp;

                        /* Right point */
                        z1 = *(pMag+1);
                        z2 = *(pMag-Column+1);

                        mag2 = (m00 - z1)*xperp + (z1 - z2)*yperp;
                    }
                    else
                    {
                        /* 100 */
                        /* Left point */
                        z1 = *(pMag+Column);
                        z2 = *(pMag+Column-1);

                        mag1 = (z1 - z2)*xperp + (m00 - z1)*yperp;

                        /* Right point */
                        z1 = *(pMag-Column);
                        z2 = *(pMag-Column+1);

                        mag2 = (z1 - z2)*xperp  + (m00 - z1)*yperp;
                    }
                }
            }
            else
            {
                if( (int)*pDy >= 0 )
                {
                    if (-(int)*pDx >= (int)*pDy)
                    {
                        /* 011 */
                        /* Left point */
                        z1 = *(pMag+1);
                        z2 = *(pMag-Column+1);

                        mag1 = (z1 - m00)*xperp + (z2 - z1)*yperp;

                        /* Right point */
                        z1 = *(pMag-1);
                        z2 = *(pMag+Column-1);

                        mag2 = (z1 - m00)*xperp + (z2 - z1)*yperp;
                    }
                    else
                    {
                        /* 010 */
                        /* Left point */
                        z1 = *(pMag-Column);
                        z2 = *(pMag-Column+1);

                        mag1 = (z2 - z1)*xperp + (z1 - m00)*yperp;

                        /* Right point */
                        z1 = *(pMag+Column);
                        z2 = *(pMag+Column-1);

                        mag2 = (z2 - z1)*xperp + (z1 - m00)*yperp;
                    }
                }
                else
                {
                    if (-(int)*pDx > -(int)*pDy)
                    {
                        /* 001 */
                        /* Left point */
                        z1 = *(pMag+1);
                        z2 = *(pMag+Column+1);

                        mag1 = (z1 - m00)*xperp + (z1 - z2)*yperp;

                        /* Right point */
                        z1 = *(pMag-1);
                        z2 = *(pMag-Column-1);

                        mag2 = (z1 - m00)*xperp + (z1 - z2)*yperp;
                    }
                    else
                    {
                        /* 000 */
                        /* Left point */
                        z1 = *(pMag+Column);
                        z2 = *(pMag+Column+1);

                        mag1 = (z2 - z1)*xperp + (m00 - z1)*yperp;

                        /* Right point */
                        z1 = Mag[i-1][j];
                        z2 = Mag[i-1][j-1];

                        mag2 = (z2 - z1)*xperp + (m00 - z1)*yperp;
                    }
                }
            }

            /* Now determine if the current point is a maximum point */

            if (((int)mag1 >= 0) || ((int)mag2 > 0))
            {
                *pTemp = NO_EDGE;
            }
            else
            {
                *pTemp = POSSIBLE_EDGE;
            }
        }
     }
}

void Edge::FollowEdges(int xs,int ys,int Tlow){
     int i,j,k ;

     for(k=0;k<8;k++){
       i = xs + x[k] ;
       j = ys + y[k] ;
       if(ValidPixel(i,j))
       if( (edge[i][j] == POSSIBLE_EDGE) && ( Mag[i][j]>= Tlow)){
         edge[i][j] = EDGE ;
         FollowEdges(i,j,Tlow);
       }
     }
}

void Edge::TwoLevelThresholding(){
   int i,j,index                                                     ;
   float low,high,ThLow,ThHigh,Mag_min,Mag_max                       ;
   bool bLow = TRUE , bHigh = TRUE                                   ;

   Mag_min = 255.0      ;
   Mag_max = 0.0        ;
   float *pMag ;
   for (i=0;i<Row;i++){
       pMag= Mag[i];
       for (j=0;j<Column;j++,pMag++){
           if (Mag_min>*pMag) Mag_min= *pMag ;
           if (Mag_max<*pMag) Mag_max= *pMag ;
       }
   }

   memset(histogram,0,sizeof(float)*257);

   for (i=0;i<Row;i++)
       for(j=0;j<Column;j++){
         index = (int) ( (256.0 * (Mag[i][j] - Mag_min))/
                         (Mag_max - Mag_min + 1.0)) ;
         index = index>256.0 ? 256.0 : index	;
         histogram[(unsigned char)index] ++ ;
      }

   for (i=0;i<257;i++)
       histogram[i] /= ((float)  Row * (float) Column);

   for(i=1;i<257;i++)
      histogram[i] += histogram[i-1] ;
   memcpy(edge[0],temp[0],sizeof(unsigned char)*Row*Column);

   low  = TLow   / 100.0 ;
   high = THigh  / 100.0 ;

   for(i=0;i<257;i++){
      if((histogram[i]>=low) && (bLow == TRUE)){
        ThLow  = ( ( (Mag_max-Mag_min + 1.0)* ((float) i - Mag_min + 1.0)) / 256.0) + Mag_min ;
        bLow = FALSE                                                                   ;
      }
      if((histogram[i]>=high) && (bHigh == TRUE)){
        ThHigh = ( ( (Mag_max-Mag_min + 1.0) * ((float) i - Mag_min + 1.0)) / 256.0) + Mag_min ;
        bHigh = FALSE ;
      }
    }

  for(i=1;i<Row-1;i++)
      for(j=1;j<Column-1;j++){
         if( (edge[i][j] == POSSIBLE_EDGE) && (Mag[i][j]>=ThHigh)){
           edge[i][j] = EDGE ;
           FollowEdges(i,j,ThLow) ; }
  }
  unsigned char *p;
  for (i=0;i<Row;i++){
      p=edge[i] ;
      for (j=0;j<Column;j++,p++)
          if (*p!=EDGE) *p= NO_EDGE ;
  }
}

void Edge::thinning()
{
  int rw, cl;
  pic = new char[Row*Column];
  memcpy(pic,edge[0],Row*Column*sizeof(char)) ;
  thinHIPS();
  memcpy(edge[0],pic,Row*Column*sizeof(char)) ;
  delete []pic;
}

void Edge::thinHIPS()
{
	int r,c,change,trans,neigh,pass,first,gap,gapopen;
	int i,j ;
	int mini,sumr,sumc,label,npt,Mfound;
	float avgr,avgc,mindist,dist;
	char *ppic;

	if (flagpass[0]==0 && flagpass[1]==0 && flagpass[2]==0 &&
		flagpass[3]==0)
		flagpass[0]=flagpass[1]=1;

	    for (pass=0;pass<4;pass++)
		if (flagpass[pass]) {

		    do {
			maxlabel = maxmult = change = 0;

			ppic = pic;

			/*  neighborhood:  432
					   501
					   678
		    	and 9 = 1 for modular arithmetic */

		    for (r=0;r<Row;r++) {
			for (c=0;c<Column;c++) {
				if ((*ppic&0377)!=0) {
				    switch(pass) {

				    case 0:
				    case 2:
					neigh = trans = 0;
					for (i=1;i<=8;i++) {
						if (neighbor(r,c,i)!=0)
							neigh++;
						if (((neighbor(r,c,i)==0) ^
						    (neighbor(r,c,i+1)==0))!=0)
							trans++;
					}
					if (pass==0) {
					    if (trans==2 && neigh>2 && neigh<6) {
						*ppic++ = 0;
						change++;
					    }
					    else
						ppic++;
					}
					else {   /* pass 2 */
					    if (trans == 2 && neigh < 3)
						*ppic++ = E;
					    else if (neigh == 0)
						*ppic++ = I;
					    else if (trans >= 6)
						*ppic++ = M;
					    else if (neigh > 2)
						*ppic++ = MM;
					    else
						*ppic++ = U;
					}
					continue;

				    case 1:
					neigh = trans = 0;
					for (i=1;i<=8;i++) {
						if (neighbor(r,c,i)) {
							neigh++;
							first = i;
						}
                                                if (((neighbor(r,c,i)==0) ^
						    (neighbor(r,c,i+1)==0))!=0)
							trans++;
					}
					if (neigh<2 || neigh>6 ||
					    (neigh==2 && trans<4) ||
					    (neigh==3 && trans>4)) {
						ppic++;
						continue;
					}
					gap = gapopen = 0;
					for (i=first+1;i<first+8;i++) {
					    j = ((i - 1) % 8) + 1;
					    if (gapopen) {
						if (neighbor(r,c,j))
							gapopen = 0;
					    }
					    else {
						if (j%2==1 && neighbor(r,c,j)==0) {
							gapopen++;
							gap++;
						}
					    }
					}
					if (gap == 1) {
						*ppic++ = 0;
						change++;
					}
					else
						ppic++;
					continue;

				    case 3:
					if ((*ppic&0377)==MM || (*ppic&0377)==M) {
					    ppic++;
					    first = 0;
					    for (i=1;i<=8;i++) {
						if (neighbor(r,c,i)==MM || neighbor(r,c,i)==M) {
						    if (first++ == 0)
							addmult(r,c,++maxlabel);
						    addmult(r+rinc[i],
							c+cinc[i],maxlabel);
						}
					    }
					}
					else
						ppic++;
					continue;

				    default:
					perror("unknown pass!"); /*IE*/
				    }
				}
				else
					ppic++;
		        }
		    }

		    if (pass == 3) {
			while (1) {
			    label = 0;
			    for (i=0;i<maxmult;i++)
				if ((label = mlabel[i]) != 0)
				    break;
			    if (label == 0)
				break;
			    Mfound = sumr = sumc = npt = 0;
			    for (i=0;i<maxmult;i++)
				if (mlabel[i] == label) {
				    sumr += mrow[i];
				    sumc += mcol[i];
				    if ((*(pic + mrow[i]*Column+mcol[i])&0377) == M)
					Mfound++;
				    npt++;
				}
			    if (Mfound) {
				for (i=0;i<maxmult;i++) {
					if (mlabel[i] == label)
						mlabel[i] = 0;
				}
				continue;
			    }
			    if (npt < 2)
				perror("only one M point!??"); /*IE*/
			    avgr = ((float) sumr)/npt;
			    avgc = ((float) sumc)/npt;
			    mindist = 10000000.;
			    for (i=0;i<maxmult;i++) {
				if (mlabel[i] == label) {
				    dist = absval(avgr-mrow[i]) +
					absval(avgc-mcol[i]);
				    if (dist<mindist) {
					mindist = dist;
					mini = i;
				    }
				}
			    }
			    for (i=0;i<maxmult;i++) {
				if (mlabel[i]==label) {
				    if (i == mini) {
					*(pic + mrow[i]*Column + mcol[i]) = M;
					change++;
				    }
				    mlabel[i] = 0;
				}
			    }
			}
		    }
		    } while (change!=0 && flags==0 && (pass==0 || pass==1));
		}
}

void Edge::addmult(int row,int col,int label)
{
	int i,oldlabel;

	oldlabel = 0;
	for (i=0;i<maxmult;i++)
		if (mrow[i]==Row && mcol[i]==col)
			oldlabel = mlabel[i];
	if (oldlabel) {
		for (i=0;i<maxmult;i++)
			if (mlabel[i] == oldlabel)
				mlabel[i] = label;
	}
	else {
		if (maxmult >= NMM)
			perror("M point overflow");
		mrow[maxmult] = row;
		mcol[maxmult] = col;
		mlabel[maxmult++] = label;
	}
}

inline int Edge::neighbor(int r,int c,int d)
{
	int i,j;

	i = r + rinc[d];
	j = c + cinc[d];
	if (i<0 || i>=Row || j<0 || j>=Column)
		return(0);
	else
		return(pic[i*Column+j]&0377);
}

void Edge::ClearBorders(){
     int i    ;

     for(i=0;i<Row;i++){
        edge[i][0]        = NO_EDGE ;
        edge[i][Column-1] = NO_EDGE ;
     }
     for(i=0;i<Column;i++){
        edge[0][i]     = NO_EDGE ;
        edge[Row-1][i] = NO_EDGE ;
     }
}

