/***************************************************************************/
/* 			     		    	                           */
/*  TWO DIMENSIONAL GENERALIZED EDGE DETECTOR           		   */
/* 			     		    	                           */
/*  B. Kurt, M. Gokmen, "Two Dimensional Generalized Edge Detector,"       */
/*          appear to the 10th International Conference on Image Analysis  */
/*                           and Processing, ICIAP'99, Venice Italy, 1999. */
/* 			     		    	                           */
/***************************************************************************/
/* 			     		    	                           */
/*  Edge Class Library       		                                   */
/* 			     		    	                           */
/*  edge              	     		    	                           */
/*    |              	     		    	                           */
/*  image              	     		    	                           */
/* 			     		    	                           */
/***************************************************************************/
/* 			     		    	                           */
/*  Last modified by    		    	                           */
/*                   Binnur KURT            	                           */
/*  When        	     		    	                           */
/* 		     August 19th 1999    	                           */
/* 			     		    	                           */
/***************************************************************************/
#include <math>

#include "image.h"

#ifndef __EDGE__
#define __EDGE__

#define CANNY                   1
#define _1D_GED                 2
#define _2D_GED                 3

#define NO_EDGE                 0
#define POSSIBLE_EDGE           128
#define EDGE                    250

#define U	010
#define E	020
#define I	040
#define MM	0100
#define M	0200
#define	NMM	1000
#define absval(f) (((f)<0)?(-(f)):(f))

int x[8] = {1,1,0,-1,-1,-1,0,1},y[8] = {0,1,1,1,0,-1,-1,-1}     ;
int flagpass[4] = {0,0,0,0}                                     ;
int flags = 0                                                   ;
int flagv = 0                                                   ;

int rinc[10] = {0,0,-1,-1,-1,0,1,1,1,0}			        ;
int cinc[10] = {0,1,1,0,-1,-1,-1,0,1,1}			        ;
int mrow[NMM],mcol[NMM],mlabel[NMM],maxmult,maxlabel	        ;
char *pic							;

class Edge:public Image{
      public:
            Edge()                                      ;
           ~Edge()                                      ;
       void SaveEdgesToTImage(TImage *)                 ;
       void SetThresholds(int,int)                      ;
       void SetLambda(float)                            ;
       void SetTau(float)                               ;
       void GetEdges()                                  ;
       void RFilter()                                   ;
       void ClearBorders()                              ;
       void SetGFilterLength(int)                       ;
       void FreeTemporary()                             ;
       unsigned char **edge                             ;
       void  FollowEdges(int,int,int)                   ;
       void  NonMaxSuppression()                        ;
       void  TwoLevelThresholding()                     ;
       void  thinning ()	                        ;
      private:
       int  FilterLength                                ;
       int  TLow,THigh                                  ;
       float Lambda,Tau                                 ;
       unsigned char **temp                             ;
       float **Dx,**Dy,**Mag                            ;
       float histogram[257]                             ;
       bool AreEdgesDetected                            ;

       void  GetGxFilter()                              ;
       void  GetGyFilter()                              ;
       void  GetRFilter()                               ;
       float Signum(int)                                ;
       void  thinHIPS()			                ;
       int   neighbor(int ,int ,int)                    ;
       void  addmult(int,int,int)                       ;
} ;
#endif

