/***************************************************************************/
/* 			     		    	                           */
/*  Image Class Library       		                                   */
/* 			     		    	                           */
/*  Image data structure and simple manipulations i.e. I/O, convolution    */
/* 			     		    	                           */
/***************************************************************************/
/* 			     		    	                           */
/*  Last modified by    		    	                           */
/*                   Binnur KURT            	                           */
/*  When        	     		    	                           */
/* 		     August 19th 1999    	                           */
/* 			     		    	                           */
/***************************************************************************/
#include <string>
#include <iostream>

using namespace std ;

#include "Image.h"

Image::Image(){
     Y= 0L;
     filterOutput= 0L;
     IsMemoryAllocatedForY= false;
     IsMemoryAllocatedForFilter= false;
     IsNewImageLoaded= false;
}

Image::~Image(){
    freeAll() ;
}

void Image::AssignName(char *name){
     filename = new char[256]	;
     strcpy(filename,name)              ;
     length = strlen(name)              ;
}

void Image::PrintName(){
     cout << filename ;
}

void Image::PasteImageFromAnotherImage(Image& TheOther){
     int i,j    ;

     if( (TheOther.Width != Width) || (TheOther.Height != Height) ){
      delete[] Y[0] ;
      delete[] Y ;
      Width  = Column = TheOther.Width ;
      Height = Row = TheOther.Height ;
      Y = new unsigned char*[Row] ;
      Y[0]= new unsigned char[Row*Column] ;
      unsigned char *p=Y[0]+Column;
      for (i=1;i<Row;i++,p+=Column)
          Y[i]= p ;
     }
     memcpy(Y[0],TheOther.Y[0],sizeof(unsigned char)*Row*Column);
}

void Image::freeAll(){
     int i ;

     if(Y != NULL){
       delete[] Y[0] ;
       delete[] Y ;
       Y= 0L;
       IsMemoryAllocatedForY = false ;
     }
     if (filterOutput!=0L){
        delete[] filterOutput[0] ;
        delete[] filterOutput ;
        filterOutput= 0l;
        IsMemoryAllocatedForFilter = false ;
     }
     if (filename) delete[] filename;
     filename= 0L;
}

void Image::LoadFromTImage(TImage *img){
     int i,j,r,g,b                      ;
     double luminance                    ;

     img->Hint = filename                               ;
     if (IsMemoryAllocatedForY == true){
        IsNewImageLoaded           = true                ;
        IsMemoryAllocatedForY      = false               ;
        IsMemoryAllocatedForFilter = false               ;
        freeAll()                                        ;
     }

     Width = Column = img->Picture->Bitmap->Width ;
     Height = Row = img->Picture->Bitmap->Height ;

     Y = new unsigned char *[Row] ;
     Y[0] = new unsigned char [Row*Column] ;
     unsigned char *p=Y[0]+Column ;
     for (i=1;i<Row;i++,p+=Column)
         Y[i] = p ;
     IsMemoryAllocatedForY = true       ;
     if (img->Picture->Bitmap->PixelFormat==pf24bit){
        for (i=0;i<Row;i++){
             unsigned char *p=Y[i] ;
             unsigned char *q = (unsigned char *) img->Picture->Bitmap->ScanLine[i] ;
             for (j=0;j<Column;j++){
                 r = *q++ ;
                 g = *q++ ;
                 b = *q++ ;
                 luminance = 0.299 * static_cast<double>(r) +
                             0.587 * static_cast<double>(g) +
                             0.114 * static_cast<double>(b) ;
                 *p++ = static_cast<unsigned char>(luminance+0.5) ;
             }
        }
     }
     else if (img->Picture->Bitmap->PixelFormat==pf8bit){
          for (i=0;i<Row;i++){
              unsigned char *p=Y[i] ;
              unsigned char *q = (unsigned char *) img->Picture->Bitmap->ScanLine[i] ;
              memcpy(p,q,sizeof(unsigned char)*Column) ;
          }
     }
     else{
       ShowMessage("Does not support this Pixel Format.") ;
       freeAll();
     }
}

void Image::SaveToTImage(TImage *img){
     static int i,j;
     unsigned char *pLine,*p ;
     img->Hint = filename;

     p=Y[0];
     for (i=0;i<Row;i++){
         pLine = (unsigned char*) img->Picture->Bitmap->ScanLine[i];
         for (j=0;j<Column;j++,p++){
             *pLine++ = *p;
             *pLine++ = *p;
             *pLine++ = *p;
         }
     }
     img->Refresh();
}

void Image::Invert(){
     int i,j                      ;
     unsigned char *p=Y[0] ;
     for (i=0;i<Row;i++)
         for (j=0;j<Column;j++,p++)
             *p = 255 - *p;
}

inline void Image::SetFilterLength(int height,int width){
     filter.Width = width ;
     filter.Height = height ;
}

inline void Image::SetFilterOrigin(int x,int y){
     filter.xx = x ;
     filter.yy = y ;
}

inline bool Image::ValidPixel(int x,int y){
    if((x>=0)&&(y>=0)&&(x<Height)&&(y<Width))
      return true ;
    else
      return false ;
}

void Image::Convolve(){
     int i,j,x,y,conPointX,conPointY;
     float conv,*p,*pFilter;
     unsigned char *q;

     if (IsMemoryAllocatedForFilter == false){
        filterOutput = new float*[Row] ;
        filterOutput[0] = new float[Row*Column] ;
        p= filterOutput[0]+Column ;
        for (i=1;i<Row;i++,p+=Column)
            filterOutput[i] = p ;
     }

     IsMemoryAllocatedForFilter= true ;

     for (i=filter.xx;i<Row-filter.xx;i++){
         p=filterOutput[i] ;
         for (j=filter.yy;j<Column-filter.yy;j++){
             conv = 0.0			        ;
             pFilter=filter.coeff[0];
             conPointX=i+filter.xx;
             for (x=0;x<filter.Height;x++,conPointX--){
                 conPointY= j+filter.yy;
                 q= Y[conPointX]+conPointY;
                 for (y=0;y<filter.Width;y++,pFilter++,conPointY--,q--){
                     if (ValidPixel(conPointX,conPointY)){
                        conv += *pFilter * (float) *q ;
                     }
                     else
                     if (ValidPixel(i-filter.xx+x,j-filter.yy+y)){
                        //conv += *pFilter * (float) Y[i-filter.xx+x][j-filter.yy+y] ;
                     }
                 }
             }
             conv *= filter.Gain ;
             *p++ = conv ;
         }
     }
}
