#include "edge.h"

typedef struct COORDINATE{
        int     x       ;
        int     y       ;
} CoordinateStruct      ;

int  CrossConversion[3][3]={{3 , 2  , 1 },
                            {4 , -1 , 0 },
                            {5 , 6  , 7 }} ;

typedef struct CONTOUR_BOUNDING_BOX{
        int BoxWidth                                    ;
        int BoxHeight                                   ;
        int BoxBeginH                                   ;
        int BoxBeginV                                   ;
        float BoxCenterH                                ;
        float BoxCenterV                                ;
} ContourBoundingBoxStruct                              ;

typedef struct CONTOUR_LIST{
        CoordinateStruct location                       ;
        int ChainCode                                   ;
        bool IsJunction                                 ;
} ContourListStruct                                     ;

typedef struct CONTOUR_DESCRIPTOR{
        bool IsClosedContour                            ;
        bool IsConvexContour                            ;
        bool IsSimpleContour                            ;
        int  ContourLength                              ;
        ContourBoundingBoxStruct BoundingBox            ;
        ContourListStruct *ContourList                  ;
} ContourDescriptorStruct                               ;

class contour:public edge{
      public:
            contour()                                   ;
           ~contour()                                   ;
      void  BlobColoring()                              ;
      void  ExtractContourDescriptor()                  ;
      void  SaveContoursToTImage(TImage *)              ;
      void  SaveDescriptorToTImage(TImage *)            ;
      void  DrawBoundingBox(TImage *)                   ;
      void  WriteContourDescriptorToFile(char *)        ;

      int **BlobColor                                   ;
      int   NumberOfContours                            ;
      int   NumberOfEdges                               ;
      ContourDescriptorStruct *ContourDescriptor        ;
      friend CoordinateStruct operator+(CoordinateStruct,CoordinateStruct)     ;

      private:
      CoordinateStruct Displacement[8]                  ;
      bool HasBlobColoringBeenDone                      ;
      bool HasContourDescriptorExtracted                ;
      CoordinateStruct GetNextEndPoint()                ;
      CoordinateStruct GetFirstPoint()                  ;
      void FollowEndPoint(CoordinateStruct,int)         ;
      void GetContourInfo(int)                          ;
      void FollowThePoint(CoordinateStruct,int,int)     ;
      CoordinateStruct FindEndPointOfTheContour(int)    ;
      CoordinateStruct FindFirstPointOfTheContour(int)  ;
} ;

     contour::contour(){

      NumberOfContours = 0                              ;
      NumberOfEdges    = 0                              ;
      HasBlobColoringBeenDone = false                   ;
      HasContourDescriptorExtracted = false             ;

      Displacement[0].x=0; Displacement[0].y=1          ;
      Displacement[1].x=-1;Displacement[1].y=1          ;
      Displacement[2].x=-1;Displacement[2].y=0          ;
      Displacement[3].x=-1;Displacement[3].y=-1         ;
      Displacement[4].x=0 ;Displacement[4].y=-1         ;
      Displacement[5].x=1 ;Displacement[5].y=-1         ;
      Displacement[6].x=1 ;Displacement[6].y=0          ;
      Displacement[7].x=1 ;Displacement[7].y=1          ;
}

contour::~contour(){

      if(HasContourDescriptorExtracted == true){
        for(k=0;k<NumberOfContours;k++)
           free(ContourDescriptor[k].ContourList) ;
           delete[] ContourDescriptor ;
           delete[] BlobColor[0] ;
           delete[] BlobColor ;
      }else
        if(HasBlobColoringBeenDone == true){
           delete[] BlobColor[0] ;
           delete[] BlobColor ;
        }
}

void  contour::ExtractContourDescriptor(){
      int i,j,k,*p;

      HasContourDescriptorExtracted= true;
      if (HasBlobColoringBeenDone==false){
         HasBlobColoringBeenDone= true;
         BlobColoring();
      }
      /* Memory allocation for ContourDescriptor  */
      ContourDescriptor = new ContourDescriptorStruct[NumberOfContours];

      for (k=0;k<NumberOfContours;k++){
          ContourDescriptor[k].BoundingBox.BoxBeginH= Row;
          ContourDescriptor[k].BoundingBox.BoxBeginV= Column;
          ContourDescriptor[k].BoundingBox.BoxWidth= -1;
          ContourDescriptor[k].BoundingBox.BoxHeight= -1;
          ContourDescriptor[k].BoundingBox.BoxCenterH= 0.0;
          ContourDescriptor[k].BoundingBox.BoxCenterV= 0.0;
          ContourDescriptor[k].ContourLength= 0;
      }
      for (i=0;i<Row;i++){
          p=BlobColor[i];
          for (j=0;j<Column;j++){
              k= *p++;
              if (k>0){
                 k--;
                 ContourDescriptor[k].ContourLength++;
                 if (i<ContourDescriptor[k].BoundingBox.BoxBeginV)
                    ContourDescriptor[k].BoundingBox.BoxBeginV= i;
                 if (j<ContourDescriptor[k].BoundingBox.BoxBeginH)
                    ContourDescriptor[k].BoundingBox.BoxBeginH= j;
                 if (i>ContourDescriptor[k].BoundingBox.BoxHeight)
                    ContourDescriptor[k].BoundingBox.BoxHeight= i;
                 if (j>ContourDescriptor[k].BoundingBox.BoxWidth)
                    ContourDescriptor[k].BoundingBox.BoxWidth= j;
                 ContourDescriptor[k].BoundingBox.BoxCenterH+= (float)j;
                 ContourDescriptor[k].BoundingBox.BoxCenterV+= (float)i;
              }
          }
      }

      for(k=0;k<NumberOfContours;k++){
       ContourDescriptor[k].BoundingBox.BoxWidth   -=
                ContourDescriptor[k].BoundingBox.BoxBeginH              ;
       ContourDescriptor[k].BoundingBox.BoxHeight  -=
                ContourDescriptor[k].BoundingBox.BoxBeginV              ;
       ContourDescriptor[k].BoundingBox.BoxCenterH /=
                (float)ContourDescriptor[k].ContourLength               ;
       ContourDescriptor[k].BoundingBox.BoxCenterV /=
                (float)ContourDescriptor[k].ContourLength               ;
       ContourDescriptor[k].ContourList = (ContourListStruct *)
                     malloc(ContourDescriptor[k].ContourLength *
                             sizeof(ContourListStruct))                 ;
      }
      for(k=1;k<=NumberOfContours;k++){
          GetContourInfo(k) ;
      }
      int *p ;
      for (i=0;i<Row;i++){
          p= BlobColor[i];
          for (j=0;j<Column;j++){
              *p++ -= NumberOfContours ;
          }
      }
}

void  contour::GetContourInfo(int contourNo){
      CoordinateStruct endPoint         ;

      endPoint = FindEndPointOfTheContour(contourNo)   ;
      if( (endPoint.x == -1) && (endPoint.y == -1)){
         ContourDescriptor[contourNo-1].IsClosedContour = true ;
         endPoint = FindFirstPointOfTheContour(contourNo) ;
      }
      else
         ContourDescriptor[contourNo-1].IsClosedContour = false ;
      FollowThePoint(endPoint,contourNo,0) ;
}

void contour::FollowThePoint(CoordinateStruct endPoint,int contourNo,int index){
     int k,l                  ;

     ContourDescriptor[contourNo-1].ContourList[index].location = endPoint ;
     BlobColor[endPoint.x][endPoint.y] += NumberOfContours ;
     for(k=-1;k<2;k++)
        for(l=-1;l<2;l++)
           if(ValidPixel(endPoint.x+k,endPoint.y+l)&&((k!=0)||(l!=0)))
             if(BlobColor[endPoint.x+k][endPoint.y+l]==contourNo){
               endPoint.x += k ;
               endPoint.y += l ;
               FollowThePoint(endPoint,contourNo,index+1) ;
               break;
             }
}

CoordinateStruct contour::FindEndPointOfTheContour(int contourNo){
     int i,j,k,l,n                              ;
     CoordinateStruct   endPoint={-1,-1}        ;

     for(i=0;i<Row;i++)
        for(j=0;j<Column;j++){ // 1
           if(BlobColor[i][j]== contourNo){ // 2
             n=0;
             for(k=-1;k<2;k++)
                for(l=-1;l<2;l++)
                   if(ValidPixel(i+k,j+l)&&((k!=0)||(l!=0)))
                     if(BlobColor[i+k][j+l]==contourNo)
                       n++;

               if(n==1){ // 3
                 endPoint.x = i         ;
                 endPoint.y = j         ;
                 return endPoint        ;
               } // 3
           } // 2
        } // 1
     return endPoint        ;
}

CoordinateStruct contour::FindFirstPointOfTheContour(int contourNo){
     int i,j                                    ;
     CoordinateStruct   endPoint={-1,-1}        ;

     for(i=0;i<Row;i++)
        for(j=0;j<Column;j++)
           if(BlobColor[i][j]== contourNo){
             endPoint.x = i         ;
             endPoint.y = j         ;
             return endPoint        ;
           }
      return endPoint        ;           
}

void  contour::BlobColoring(){
      int i,j;
      CoordinateStruct endPoint ;

      BlobColor = new int *[Row];
      BlobColor[0] = new int  [Row*Column];
      int *p=BlobColor[0]+Column;
      for (i=1;i<Row;i++,p+=Column)
          BlobColor[i]= p;

      HasBlobColoringBeenDone = true    ;
      unsigned char *pEdge;
      for (i=0;i<Row;i++){
          pEdge= edge[i];
          p= BlobColor[i];
          for (j=0;j<Column;j++,p++,pEdge++){
              if (*pEdge){
                 *p= -1;
                 NumberOfEdges++;
              }else *p=0;
          }
      }
          
      while(NumberOfEdges){
        endPoint = GetNextEndPoint()    ;
        if((endPoint.x==-1)&&(endPoint.y==-1)) break;
        NumberOfContours++;
        FollowEndPoint(endPoint,0) ;
      }

       do{
        endPoint = GetFirstPoint()    ;
        if((endPoint.x==-1)&&(endPoint.y==-1)) break;
        NumberOfContours++;
        FollowEndPoint(endPoint,1) ;
      }while(1);
}

void contour::FollowEndPoint(CoordinateStruct endPoint,
                             int        predictedDirection){
     int k,l,m,n,xx,yy                  ;
     int direction                      ;
     CoordinateStruct predictedPoint    ;

     BlobColor[endPoint.x][endPoint.y] = NumberOfContours               ;

     n=0;
     for(k=-1;k<2;k++)
        for(l=-1;l<2;l++)
           if(ValidPixel(endPoint.x+k,endPoint.y+l)&&((k!=0)||(l!=0)))
             if(BlobColor[endPoint.x+k][endPoint.y+l]==-1){
               ++n;
               if(n==1){xx=k;yy=l;}
             }
     if(n>0)
     if(n==1){ /* no alternatives to visit */
        predictedPoint.x = endPoint.x + xx              ;
        predictedPoint.y = endPoint.y + yy              ;
        predictedDirection = CrossConversion[xx+1][yy+1]        ;
        FollowEndPoint(predictedPoint,predictedDirection)  ;
     }else{
       if(n>1){ /* in this case choose the best direction */
         predictedPoint = endPoint + Displacement[predictedDirection]        ;
         if(ValidPixel(predictedPoint.x,predictedPoint.y))
           if(BlobColor[predictedPoint.x][predictedPoint.y]==-1){
             FollowEndPoint(predictedPoint,predictedDirection)  ;
           }
           else{
             m=1;
             do{
             direction = (predictedDirection + m + 8)%8       ;
             predictedPoint = endPoint + Displacement[direction]   ;
             if(ValidPixel(predictedPoint.x,predictedPoint.y))
               if(BlobColor[predictedPoint.x][predictedPoint.y]==-1){
                 FollowEndPoint(predictedPoint,direction)  ;
                 break;
               }
             if(m<0) m=-m+1     ;
             else m = -m        ;
             }while(m!=4);
           } /* else */
       }
     }
}

CoordinateStruct contour::GetNextEndPoint(){
     int i,j,k,l,n                              ;
     CoordinateStruct   endPoint={-1,-1}        ;

     for(i=0;i<Row;i++)
        for(j=0;j<Column;j++){ // 1
           if(BlobColor[i][j]==-1){ // 2
             n=0;
             for(k=-1;k<2;k++)
                for(l=-1;l<2;l++)
                   if(ValidPixel(i+k,j+l)&&((k!=0)||(l!=0)))
                     if(BlobColor[i+k][j+l]==-1)
                       n++;
             if(n==0)
               BlobColor[i][j]=0        ;
             else
               if(n==1){ // 3
                 endPoint.x = i         ;
                 endPoint.y = j         ;
                 return endPoint        ;
               } // 3
           } // 2
        } // 1
     return endPoint        ;
}

CoordinateStruct contour::GetFirstPoint(){
     int i,j                                    ;
     CoordinateStruct firstPoint={-1,-1}        ;

     for(i=0;i<Row;i++)
        for(j=0;j<Column;j++)
           if(BlobColor[i][j]==-1){
              firstPoint.x = i                  ;
              firstPoint.y = j                  ;
              return firstPoint                 ;
           }
     return firstPoint                          ;
}

CoordinateStruct operator+(CoordinateStruct a,CoordinateStruct b){
          CoordinateStruct c       ;

          c.x = a.x + b.x       ;
          c.y = a.y + b.y       ;
          return c      ;
}

void contour::WriteContourDescriptorToFile(char *filename){
     FILE *hFile        ;
     int  k             ;
     hFile = fopen(filename,"w+")       ;
     if(hFile!=NULL)
     for(k=0;k<ContourDescriptor[k].ContourLength;k++)
        fwrite(ContourDescriptor,sizeof(ContourDescriptorStruct),NumberOfContours,hFile) ;
     fclose(hFile)      ;
}

void contour::SaveContoursToTImage(TImage *img){
     int i,j                      ;

     for(i=0;i<Row;i++)
        for(j=0;j<Column;j++)
            img->Canvas->Pixels[j][i] = (TColor) (BlobColor[i][j] * 6789) ;
}

void contour::SaveDescriptorToTImage(TImage *img){
     int i,j,k,l                        ;
     TRect r                             ;

     r.Left=0;
     r.Top=0;
     r.Right=Column;
     r.Bottom=Row;
     img->Canvas->Brush->Style = bsSolid                        ;
     img->Canvas->Brush->Color = clBlack                        ;
     img->Canvas->FillRect(r)                                   ;

     for(k=0;k<NumberOfContours;k++)
        for(l=0;l<ContourDescriptor[k].ContourLength;l++){
            i = ContourDescriptor[k].ContourList[l].location.x  ;
            j = ContourDescriptor[k].ContourList[l].location.y  ;
            img->Canvas->Pixels[j][i] = (TColor) ((k+1) * 6789) ;
     }
}

void  contour::DrawBoundingBox(TImage *img){
      int k     ;
      TRect myRect;

      img->Canvas->Brush->Style = bsSolid       ;

      for(k=0;k<NumberOfContours;k++){
        img->Canvas->Brush->Color = (TColor) ((k+1) * 6789)      ;
    	myRect.Top = ContourDescriptor[k].BoundingBox.BoxBeginV-1;
    	myRect.Left = ContourDescriptor[k].BoundingBox.BoxBeginH-1;
    	myRect.Bottom = ContourDescriptor[k].BoundingBox.BoxBeginV+
                        ContourDescriptor[k].BoundingBox.BoxHeight+1;
    	myRect.Right =  ContourDescriptor[k].BoundingBox.BoxBeginH+
                        ContourDescriptor[k].BoundingBox.BoxWidth+1;
    	img->Canvas->FrameRect(myRect);
      }
}
