/*******************************************************************/
/* Extracting optimal centipede model parameters                   */
/*******************************************************************/
void  Centipede::ExtractOptimalCentipedeModelParameters(){
      int k,m                                                                 ;

      /*************************************************************/
      /* for each contour                                          */
      /*    for each edgel of that contour                         */
      /*************************************************************/
      GetTheDistance()                                                        ;
      for(k=0;k<NumberOfContours;k++){
         for(m=0;m<ContourDescriptor[k].ContourLength;m++){
             GetTheDataToBeModelled(k,m)                                      ;
             distLeft  = ContourDescriptor[k].ContourList[m].distLeft         ;
             distRight = ContourDescriptor[k].ContourList[m].distRight        ;
//           ApplyMedianFilterToTheData(5)                                    ;
             SearchForTheOptimalModelParameters(k,m)                          ;
            }
      }
}

void  Centipede::GetTheDistance(void){
      int i,k,m                                                               ;
      short int *tmpDistanceLeftArray                                         ;
      short int *tmpDistanceRightArray                                        ;
      short middle,smallest,largest                                           ;
      char dualNormalVect,normalvector                                        ;
      CoordinateStruct p1                                                     ;

      tmpDistanceLeftArray = (short int *)
             malloc(MaximumContourLength * sizeof(short int))                 ;
      tmpDistanceRightArray = (short int *)
             malloc(MaximumContourLength * sizeof(short int))                 ;

      for(k=0;k<NumberOfContours;k++){
         for(m=0;m<ContourDescriptor[k].ContourLength;m++){
         p1 = ContourDescriptor[k].ContourList[m].location                    ;
         normalvector = ContourDescriptor[k].ContourList[m].normalVect        ;
         dualNormalVect = (normalvector+4)%8                                  ;
         i = 0 ;
         do{
           p1 = p1 + Displacement[normalvector]                               ;
           if(ValidPixel(p1.x,p1.y))
             if(BlobColor[p1.x][p1.y]<=0){
               Fn[i] = Y[p1.x][p1.y]                                          ;
             if(i>0)
               if(Fn[i] == Fn[i-1])
               break                                                          ;
             i++                                                              ;
            }
            else{
             i--                                                              ;
             i = i / 2                                                        ;
             break                                                            ;
            }
         else{
             i--                                                              ;
           break                                                              ;
         }
         }while(i< 1) ;
         tmpDistanceLeftArray[m] = i + 1                                      ;
         distLeft                = i + 1                                      ;
         /* Now sampling the center edgel */
         p1 = ContourDescriptor[k].ContourList[m].location                    ;
         Fn[distLeft] =  Y[p1.x][p1.y]                                        ;
         /* Now sampling along the direction of the dual normal vector */
         i=0                                                                  ;
         do{
           p1 = p1 + Displacement[dualNormalVect]                             ;
           if(ValidPixel(p1.x,p1.y))
             if(BlobColor[p1.x][p1.y]<=0){
               Fn[i+distLeft+1] = Y[p1.x][p1.y]                               ;
               if(i>0)
                 if(Fn[i+distLeft+1] == Fn[i+distLeft])
                   break                                                      ;
                i++                                                           ;
             }
             else{
               i--                                                            ;
               i = i / 2                                                      ;
               break                                                          ;
             }
           else{
             i--                                                              ;
             break                                                            ;
           }
         }while(i< 1)                                                      ;
         tmpDistanceRightArray[m] = i + 1                                     ;
        }
        /* Now smoothing the distanceLeft and distanceRigth */
        for(m=0;m<ContourDescriptor[k].ContourLength;m++){
            middle = tmpDistanceLeftArray[m]                                  ;
            if(m>0)
              smallest = tmpDistanceLeftArray[m-1]                            ;
            else
              smallest = tmpDistanceLeftArray[m]                              ;
            if(m<ContourDescriptor[k].ContourLength-1)
              largest = tmpDistanceLeftArray[m+1]                             ;
            else
              largest = tmpDistanceLeftArray[m]                               ;

            if( ((smallest >= middle) && (smallest <= largest)) ||
                ((smallest <= middle) && (smallest >= largest)) )
              ContourDescriptor[k].ContourList[m].distLeft = smallest         ;
            else
            if( ((middle >= smallest) && (middle <= largest)) ||
                ((middle <= smallest) && (middle >= largest)) )
              ContourDescriptor[k].ContourList[m].distLeft = middle           ;
            else
            if( ((largest >= middle) && (largest <= smallest)) ||
                ((largest <= middle) && (largest >= smallest)) )
              ContourDescriptor[k].ContourList[m].distLeft = largest          ;
        }

        for(m=0;m<ContourDescriptor[k].ContourLength;m++){
            middle = tmpDistanceRightArray[m]                                 ;
            if(m>0)
              smallest = tmpDistanceRightArray[m-1]                           ;
            else
              smallest = tmpDistanceRightArray[m]                             ;
            if(m<ContourDescriptor[k].ContourLength-1)
              largest = tmpDistanceRightArray[m+1]                            ;
            else
              largest = tmpDistanceRightArray[m]                              ;

            if( ((smallest >= middle) && (smallest <= largest)) ||
                ((smallest <= middle) && (smallest >= largest)) )
              ContourDescriptor[k].ContourList[m].distRight = smallest        ;
            else
            if( ((middle >= smallest) && (middle <= largest)) ||
                ((middle <= smallest) && (middle >= largest)) )
              ContourDescriptor[k].ContourList[m].distRight = middle          ;
            else
            if( ((largest >= middle) && (largest <= smallest)) ||
                ((largest <= middle) && (largest >= smallest)) )
              ContourDescriptor[k].ContourList[m].distRight = largest         ;
        }
      }
      free(tmpDistanceLeftArray)                                              ;
      free(tmpDistanceRightArray)                                             ;
}

void  Centipede::SearchForTheOptimalModelParameters(int c1,int c2){
      int n,m,xnPoint=4                                                       ;
      int X1min=200,X2min=200,i,j                                             ;
      float Mmin=98765.0                                                      ;
      CoordinateStruct p                                                      ;

      distLeft = ContourDescriptor[c1].ContourList[c2].distLeft               ;
      distRight = ContourDescriptor[c1].ContourList[c2].distRight             ;

      if(distLeft == 0)
         distLeft = 1 ;
      if(distRight == 0)
         distRight = 1 ;

      X[1] = distLeft 	                                                      ;

      for(m=0;m<=distLeft-1;m++){
      	 for(n=distLeft+1;n<=(distLeft+distRight);n++){
            X[0] = m                                                          ;
            X[2] = n 				                              ;
            ErrorSurface[distLeft-X[0]-1][X[2]-distLeft-1] = 0.0	      ;

            for(i=0;i<xnPoint-1;i++)
               for(j=0;j<xnPoint-1;j++)
	          G[i][j] = fabs((float)X[i] - (float)X[j])	              ;

            for(i=0;i<xnPoint-1;i++)
	       G[i][i] = LambdaForData                	                      ;

            for(i=0;i<xnPoint-1;i++) {
               G[i][xnPoint-1] = 1.0				              ;
	       d[i]            = Fn[X[i]]			              ;
            }

            for(i=0;i<xnPoint-1;i++)
	       G[xnPoint-1][i] = 1.0				              ;

            for(i=0;i<xnPoint-1;i++)
	       G[i][xnPoint-1] = 1.0				              ;

           G[xnPoint-1][xnPoint-1] = 0.0			              ;
           d[xnPoint-1]  	   = 0.0			              ;

           Solve(xnPoint)					              ;

           for(i=0;i<=distLeft+distRight;i++){
              Gn[i]  = c[0] * fabs((float)i - (float)X[0]) 	              ;
	      Gn[i] += c[1] * fabs((float)i - (float)X[1])                    ;
	      Gn[i] += c[2] * fabs((float)i - (float)X[2]) + c[3]	      ;

	      ErrorSurface[distLeft-X[0]-1][X[2]-distLeft-1] +=
                      (Fn[i]-Gn[i]) * (Fn[i]-Gn[i]) ;
           }

	   ErrorSurface[distLeft-X[0]-1][X[2]-distLeft-1] /=
                      (float) (distRight+distLeft+1)			      ;

           if( (m==0) && (n==(distLeft+1)) ){
             for(i=0;i<xnPoint;i++)
	        optimal_c[i] = c[i]				              ;
    	     Mmin = ErrorSurface[distLeft-X[0]-1][X[2]-distLeft-1]	      ;
	     X1min = X[0]				                      ;
	     X2min = X[2]            		                              ;
           }

          if(ErrorSurface[distLeft-X[0]-1][X[2]-distLeft-1]<=Mmin){
            for(i=0;i<xnPoint;i++)
	       optimal_c[i] = c[i]			                ;
            Mmin = ErrorSurface[distLeft-X[0]-1][X[2]-distLeft-1]	;
            X1min = X[0]			                        ;
            X2min = X[2]               	                                ;
          }
         }
        }

       /* Optimal Sampling Done.*/
       for(i=0;i<=distLeft+distRight;i++){
	   Gn[i]  = optimal_c[0] *  fabs((float)i - (float)X1min) 	;
	   Gn[i] += optimal_c[1] *  fabs((float)i - (float)X[1])	;
	   Gn[i] += optimal_c[2] *  fabs((float)i - (float)X2min) +
                    optimal_c[3]	                                ;

	if(Gn[i]>255.0) Gn[i]=255.0	                                ;
	if(Gn[i]<0.0)   Gn[i]=0.0		                        ;
       }

       p = ContourDescriptor[c1].ContourList[c2].location               ;

       ContourDescriptor[c1].ContourList[c2].WLeft  = distLeft-X1min    ;
       ContourDescriptor[c1].ContourList[c2].WRight = X2min-distLeft    ;
       ContourDescriptor[c1].ContourList[c2].CLeft  =
                                         Y[p.x][p.y] - (int)Gn[X1min]   ;
       ContourDescriptor[c1].ContourList[c2].CRight =
                                         Y[p.x][p.y] - (int)Gn[X2min]   ;

       for(i=0;i<4;i++)
          ContourDescriptor[c1].ContourList[c2].OptimalModelParameters[i]
                                              = optimal_c[i] ;
}

void Centipede::Solve(int N)
{
     int i,j,k		;
     float pivot	;

     for(i=0;i<N-1;i++)
	for(j=i+1;j<N;j++){
          if(G[i][i] != 0.0)
	    pivot = -G[j][i]/G[i][i]		;
          else
	    pivot = 0.0		                ;
	  G[j][i] = 0.0				;
	    for(k=i+1;k<N;k++)
	       G[j][k] += pivot * G[i][k]	;
	    d[j] += pivot * d[i] 	  	;
        }

    if(G[N-1][N-1] != 0.0)
      c[N-1] = d[N-1] / G[N-1][N-1]		;
    else
      c[N-1] = 0.0                              ;

     for(k=N-2;k>=0;k--) {
        c[k] = d[k]                             ;
        for(i=N-1;i>k;i--)
           c[k] -= G[k][i] * c[i]               ;

       if(G[k][k] != 0.0)
           c[k] /= G[k][k]                         ;
       else
           c[k] = 0.0                           ;
    }
}

void  Centipede::GetTheDataToBeModelled(int k,int m){
      char dualNormalVect,normalvector          ;
      CoordinateStruct p1                       ;
      int i                                     ;
      float swap                                ;

      distLeft = ContourDescriptor[k].ContourList[m].distLeft           ;
      distRight = ContourDescriptor[k].ContourList[m].distRight         ;
      normalvector = ContourDescriptor[k].ContourList[m].normalVect     ;
      dualNormalVect = (normalvector+4)%8                               ;
      i=0                                                               ;
      p1 = ContourDescriptor[k].ContourList[m].location                 ;
      do{
         p1 = p1 + Displacement[normalvector]                               ;
         if(ValidPixel(p1.x,p1.y))
           if(BlobColor[p1.x][p1.y]<=0){
             Fn[i] = Y[p1.x][p1.y]                                          ;
           if(i>0)
             if(Fn[i] == Fn[i-1])
             break                                                          ;
           i++                                                              ;
          }
          else{
           i--                                                              ;
           break                                                            ;
          }
       else{
         i--                                                                ;
         break                                                              ;
       }
      }while(i< distLeft);
      if(i==distLeft)
        distLeft = ContourDescriptor[k].ContourList[m].distLeft = i  ;
      else
        distLeft = ContourDescriptor[k].ContourList[m].distLeft = i+1  ;
      for(i=0;i<distLeft/2;i++){
         swap  = Fn[i]                                                  ;
         Fn[i] = Fn[distLeft-1-i]                                       ;
         Fn[distLeft-1-i] = swap                                        ;
      }
      /* Now sampling the center edgel */
      p1 = ContourDescriptor[k].ContourList[m].location                 ;
      Fn[distLeft] =  Y[p1.x][p1.y]                                     ;
      /* Now sampling along the direction of the dual normal vector */
      i=0                                                               ;
      do{
         p1 = p1 + Displacement[dualNormalVect]                         ;
         if(ValidPixel(p1.x,p1.y))
           if(BlobColor[p1.x][p1.y]<=0){
             Fn[i+distLeft+1] = Y[p1.x][p1.y]                           ;
             if(i>0)
               if(Fn[i+distLeft+1] == Fn[i+distLeft])
                 break                                                  ;
              i++                                                       ;
           }
           else{
             i--                                                        ;
             break                                                      ;
           }
         else{
           i--                                                          ;
           break                                                        ;
         }
      }while(i< distRight);
      if(i==distRight)
        distRight = ContourDescriptor[k].ContourList[m].distRight = i  ;
      else
        distRight = ContourDescriptor[k].ContourList[m].distRight = i+1  ;
}

void  Centipede::ApplyMedianFilterToTheData(int Ln){
      int i,j,k,Length,cnt      ;
      float temp[65]            ;

      Length = distLeft + distRight + 1 ;
      for(i=Ln/2;i<Length-Ln/2;i++)
         temp[i] = Fn[i] ;

      for(i=Ln/2;i<Length-Ln/2;i++){
         for(j=-Ln/2;j<=Ln/2;j++){
            cnt = 0        ;
            for(k=-Ln/2;k<=Ln/2;k++)
               if(Fn[i+j]>Fn[i+k]) cnt++ ;
            if(cnt == Ln/2){
              temp[i] = Fn[i+j] ;
              break     ;
            }
         }
      }

      for(i=Ln/2;i<Length-Ln/2;i++)
         Fn[i] = temp[i] ;
}

void Centipede::SaveModelToMatrices(unsigned char **betha,
                                    unsigned char **depth,
                                    bool  boolReconstrucTheEdgeProfile){
     int i,j,k,l,m,n,s,dataLength,dataLeft,dataRight    ;
     char normaldirection,dualnormaldirection           ;
     float c0,c1,c2,c3                                  ;
     int   x0,x1,x2                                     ;

     if(boolReconstrucTheEdgeProfile == true){
       /* Optimal Sampling Done.                                        */
        for(k=0;k<NumberOfContours;k++)
           if(ContourDescriptor[k].IsActive == true)
           for(l=0;l<ContourDescriptor[k].ContourLength;l++){
            c0 = ContourDescriptor[k].ContourList[l].OptimalModelParameters[0] ;
            c1 = ContourDescriptor[k].ContourList[l].OptimalModelParameters[1] ;
            c2 = ContourDescriptor[k].ContourList[l].OptimalModelParameters[2] ;
            c3 = ContourDescriptor[k].ContourList[l].OptimalModelParameters[3] ;

            dataLeft   = ContourDescriptor[k].ContourList[l].distLeft    ;
            dataRight  = ContourDescriptor[k].ContourList[l].distRight   ;
            dataLength = dataRight + dataLeft + 1  ;
            x0 = dataLeft - ContourDescriptor[k].ContourList[l].WLeft    ;
            x1 = dataLeft                                                ;
            x2 = dataLeft + ContourDescriptor[k].ContourList[l].WRight   ;

            normaldirection = ContourDescriptor[k].ContourList[l].normalVect ;
            dualnormaldirection = ( normaldirection + 4 ) % 8         ;
            m  = ContourDescriptor[k].ContourList[l].location.x       ;
            n = ContourDescriptor[k].ContourList[l].location.y        ;

            for(s=x0;s<=x2;s++){
               Gn[s]  = c0 *  fabs((float)s - (float)x0)	;
               Gn[s] += c1 *  fabs((float)s - (float)x1)	;
               Gn[s] += c2 *  fabs((float)s - (float)x2) + c3	;

               if(Gn[s]>255.0) Gn[s]=255.0                            ;
               if(Gn[s]<0.0)   Gn[s]=0.0		                ;
               if(s<dataLeft){
                 i = m + (dataLeft - s) * Displacement[normaldirection].x    ;
                 j = n + (dataLeft - s) * Displacement[normaldirection].y    ;
               }
               else if(s==dataLeft){
                 i = m                                                ;
                 j = n                                                ;
               }
               else{
                i = m + (s - dataLeft) * Displacement[dualnormaldirection].x ;
                j = n + (s - dataLeft) * Displacement[dualnormaldirection].y ;
               }
               if(betha[i][j] == 255)
                 break ;
               depth[i][j] = Gn[s]       ;
               betha[i][j] = 255         ;
            }
          }
     }else{
     /* the model itself.                                               */
     for(k=0;k<NumberOfContours;k++)
        if(ContourDescriptor[k].IsActive == true)     
        for(l=0;l<ContourDescriptor[k].ContourLength;l++){
            if(ContourDescriptor[k].ContourLength>1){
            normaldirection = ContourDescriptor[k].ContourList[l].normalVect ;
            dualnormaldirection = ( normaldirection + 4 ) % 8 ;
            m  = ContourDescriptor[k].ContourList[l].location.x  ;
            i  = m + ContourDescriptor[k].ContourList[l].WLeft *
                    Displacement[normaldirection].x ;
            n = ContourDescriptor[k].ContourList[l].location.y  ;
            j = n + ContourDescriptor[k].ContourList[l].WLeft *
                    Displacement[normaldirection].y ;

            depth[i][j] = Y[i][j] - ContourDescriptor[k].ContourList[l].CLeft;
            betha[i][j] = 255     ;

            depth[m][n] = Y[m][n] ;
            betha[m][n] = 255     ;

            m  = ContourDescriptor[k].ContourList[l].location.x  ;
            i  = m + ContourDescriptor[k].ContourList[l].WRight *
                    Displacement[dualnormaldirection].x ;
            n = ContourDescriptor[k].ContourList[l].location.y  ;
            j = n + ContourDescriptor[k].ContourList[l].WRight *
                    Displacement[dualnormaldirection].y ;
            depth[i][j] = Y[i][j] - ContourDescriptor[k].ContourList[l].CRight;
            betha[i][j] = 255     ;
        }
      }
    }
}

void  Centipede::SaveModelToTImage(TImage *img1,TImage *img2,
                                   bool  boolReconstrucTheEdgeProfile){
     int i,j,k,l,m,n,s,dataLength,dataLeft,dataRight    ;
     char normaldirection,dualnormaldirection           ;
     float c0,c1,c2,c3                                  ;
     TRect r                                            ;

     img2->Width = Width                        ;
     img2->Height = Height                        ;
     img2->Left = Width + 16                       ;

     r.Left=0;
     r.Top=0;
     r.Right=Column;
     r.Bottom=Row;
     img1->Canvas->Brush->Style = bsSolid                        ;
     img1->Canvas->Brush->Color = clBlack                        ;
     img1->Canvas->FillRect(r)                                   ;
     img1->Hint = "Depth data"                                   ;

     img2->Canvas->Brush->Style = bsSolid                        ;
     img2->Canvas->Brush->Color = clBlack                        ;
     img2->Canvas->FillRect(r)                                   ;
     img2->Hint = "Sparseness data"                              ;

     if(boolReconstrucTheEdgeProfile == true){
       /* Optimal Sampling Done.                                        */
        for(k=0;k<NumberOfContours;k++)
           if(ContourDescriptor[k].IsActive == true)
           for(l=0;l<ContourDescriptor[k].ContourLength;l++){
            c0 = ContourDescriptor[k].ContourList[l].OptimalModelParameters[0] ;
            c1 = ContourDescriptor[k].ContourList[l].OptimalModelParameters[1] ;
            c2 = ContourDescriptor[k].ContourList[l].OptimalModelParameters[2] ;
            c3 = ContourDescriptor[k].ContourList[l].OptimalModelParameters[3] ;
              dataLeft   = ContourDescriptor[k].ContourList[l].WLeft    ;
              dataRight  = ContourDescriptor[k].ContourList[l].distLeft +
                            ContourDescriptor[k].ContourList[l].WRight + 1 ;
              dataLength = ContourDescriptor[k].ContourList[l].distLeft +
                           ContourDescriptor[k].ContourList[l].distRight + 1 ;

              normaldirection = ContourDescriptor[k].ContourList[l].normalVect ;
              dualnormaldirection = ( normaldirection + 4 ) % 8         ;
              m  = ContourDescriptor[k].ContourList[l].location.x       ;
              n = ContourDescriptor[k].ContourList[l].location.y        ;

              for(s=0;s<dataLength;s++){
	         Gn[s]  = c0 *  fabs((float)s - (float)dataLeft) ;
	         Gn[s] += c1 *  fabs((float)s - (float)dataLeft)	;
	         Gn[s] += c2 *  fabs((float)s - (float)dataRight) + c3	;

  	         if(Gn[s]>255.0) Gn[s]=255.0                            ;
	         if(Gn[s]<0.0)   Gn[s]=0.0		                ;
                 if(s<ContourDescriptor[k].ContourList[l].distLeft){
                   i = m + (dataLeft - s) * Displacement[normaldirection].x    ;
                   j = n + (dataLeft - s) * Displacement[normaldirection].y    ;
                 }
                 else if(s==ContourDescriptor[k].ContourList[l].distLeft){
                   i = m                                                ;
                   j = n                                                ;
                 }
                 else{
                  i = m + (s - dataLeft) * Displacement[dualnormaldirection].x ;
                  j = n + (s - dataLeft) * Displacement[dualnormaldirection].y ;
                 }
                 img1->Canvas->Pixels[j][i] =  (TColor) (Gn[s] * 65793)      ;
                 img2->Canvas->Pixels[j][i] =  (TColor) ((k+1) * 6789)       ;
              }
          }
     }else{
     /* the model itself.                                               */
     for(k=0;k<NumberOfContours;k++)
        if(ContourDescriptor[k].IsActive == true)     
        for(l=0;l<ContourDescriptor[k].ContourLength;l++){
            normaldirection = ContourDescriptor[k].ContourList[l].normalVect ;
            dualnormaldirection = ( normaldirection + 4 ) % 8 ;
            m  = ContourDescriptor[k].ContourList[l].location.x  ;
            i  = m + ContourDescriptor[k].ContourList[l].WLeft *
                    Displacement[normaldirection].x ;
            n = ContourDescriptor[k].ContourList[l].location.y  ;
            j = n + ContourDescriptor[k].ContourList[l].WLeft *
                    Displacement[normaldirection].y ;

            img1->Canvas->Pixels[j][i] =  (TColor)
            ( (Y[i][j] - ContourDescriptor[k].ContourList[l].CLeft)* 65793)   ;
            img2->Canvas->Pixels[j][i] =  (TColor) ((k+1) * 6789)       ;

            img1->Canvas->Pixels[n][m] =  (TColor) (Y[m][n] * 65793 ) ;
            img2->Canvas->Pixels[n][m] =  (TColor) ((k+1) * 6789)       ;

            m  = ContourDescriptor[k].ContourList[l].location.x  ;
            i  = m + ContourDescriptor[k].ContourList[l].WRight *
                    Displacement[dualnormaldirection].x ;
            n = ContourDescriptor[k].ContourList[l].location.y  ;
            j = n + ContourDescriptor[k].ContourList[l].WRight *
                    Displacement[dualnormaldirection].y ;
            img1->Canvas->Pixels[j][i] =  (TColor)
            ( (Y[i][j] - ContourDescriptor[k].ContourList[l].CRight)* 65793)   ;
            img2->Canvas->Pixels[j][i] =  (TColor) ((k+1) * 6789)       ;
        }
     }
}
