/***************************************************************************/
/* Surface Reconstruction By Hybrid Model				   */
/* 			     ( Membrane & Thin-plate )		    	   */
/***************************************************************************/
#include "imgdisk.c"

#include    <math.h>

#define TRUE	1
#define FALSE	0

unsigned char     **d,**beta,**image,**timage  		;
float		  **u					;
float tau,lambda,error,d_error,abs_error,w,T,uc 	;
int   rows,cols,i,j            			 	;

void main(int argc,char *argv[])
{
    int   	  IN=0 						;
    unsigned char bInterpolate,bDense=FALSE			;
    float average,_lambda_3_plus_8_tau				;
    float _4_lambda_1_plus_4_tau,_lambda_1_plus_7_tau		;
    float _2_lambda_tau,_lambda_tau,_2_lambda_1_plus_tau	;
    float _lambda_1_plus_3_tau,_lambda_1_plus_5_tau		;
    float _lambda_4_plus_14_tau,_lambda_4_plus_15_tau		;
    float _lambda_3_plus_6_tau,_lambda_3_plus_7_tau		;

    if(argc<8)
      {
       printf("\n Usage : ");
       printf("\n fsor (depth) (r) [(beta),-dense] (lambda) (tau) (w) (desired error) [-i (map image)]");
       printf("\n -i  Interpolation\n");
       exit(1);
      }

    lambda       = atof(argv[4])				;
    tau          = atof(argv[5])				;
    w            = atof(argv[6])				;
    d_error      = atof(argv[7])				;
    bInterpolate = FALSE 					;
    if(argc>=9)
      if(strcmp(argv[8],"-i")==0)
	bInterpolate = TRUE 					;
    if(strcmp(argv[3],"-dense")==0)
       bDense = TRUE 						;

    printf("\n ==========================================")	;
    printf("\n = Surface Reconstruction By Hybrid Model =")	;
    printf("\n ==========================================")	;
    printf("\n Lambda     = %f",lambda)				;
    printf("\n Tau        = %f",tau)				;
    printf("\n w          = %f",w)				;
    printf("\n reading DEPTH[ %s ] image ...",argv[1])		;
    d       = readhips(argv[1],&d,&rows,&cols)         	     	;
    printf("done.")						;
    printf("\n Image Size = [ %d x %d ]",rows,cols)		;

    if(bInterpolate==TRUE)
      printf("\n [x] Interpolation");
    else
      printf("\n [ ] Interpolation");

    if(bDense==TRUE){
      printf("\n [x] Dense Data");
      beta = alloc_2Darray(rows,cols,sizeof(unsigned char))		; 
      for(i=0;i<rows;i++)
	 for(j=0;j<cols;j++)
            beta[i][j] = 1.0						;}
    else {
      printf("\n [x] Sparse Data");
    printf("\n reading sparseness[ %s ][depth] image ...",argv[3])   	;
    beta = readhips(argv[3],&beta,&rows,&cols)		     	     	;
    printf("done.\n")							;}

    u       = f_alloc_2Darray(rows,cols,sizeof(float))			;
    if(bInterpolate == TRUE)
      readhips(argv[9],&timage,&rows,&cols)				; 

    for(i=0;i<rows;i++)
	for(j=0;j<cols;j++)
	   {
	    if(beta[i][j]>0)
	      beta[i][j] = 1 ;
            if(beta[i][j] == 0)
             u[i][j] = 0.0  ;
            else
       	     u[i][j]=(float) d[i][j] ;
	   }

    _4_lambda_1_plus_4_tau = 4.0 * lambda * ( 1.0 + 4.0 * tau )   	;
    _lambda_3_plus_7_tau   = lambda * (3.0 + 7.0 * tau ) 	  	;
    _lambda_1_plus_7_tau   = lambda * (1.0 + 7.0 * tau ) 	  	;
    _lambda_1_plus_5_tau   = lambda * (1.0 + 5.0 * tau ) 	  	;
    _lambda_1_plus_3_tau   = lambda * (1.0 + 3.0 * tau ) 	  	;
    _lambda_4_plus_14_tau  = lambda * (4.0 +14.0 * tau ) 	  	;
    _lambda_3_plus_8_tau   = lambda * (3.0 + 8.0 * tau ) 	  	;
    _lambda_3_plus_6_tau   = lambda * (3.0 + 6.0 * tau ) 	  	;
    _lambda_4_plus_15_tau  = lambda * (4.0 +15.0 * tau ) 	  	;
    _2_lambda_1_plus_tau   = 2.0 * lambda * ( 1.0 + tau ) 	  	;
    _lambda_tau	           = lambda * tau			       	;
    _2_lambda_tau	   = 2.0* lambda * tau		  		;

    do
      {
       error=0.0 ;
       for(i=0;i<rows;i++)
       for(j=0;j<cols;j++)
	  {
	   uc=u[i][j] ;
	   if(( (bInterpolate==TRUE)&&(timage[i][j]==0)) || (bInterpolate==FALSE))
	   if((i>1)&&(i<rows-2)&&(j>1)&&(j<cols-2))
	     {
	      T=(float)beta[i][j]+_4_lambda_1_plus_4_tau ;
		 u[i][j]=u[i][j] - (w/T) *
			    ( T * u[i][j] -
			     _lambda_1_plus_7_tau  * ( u[i-1][j] + u[i][j-1] + u[i+1][j] + u[i][j+1] ) +
			     _2_lambda_tau *( u[i-1][j-1] +u[i-1][j+1] +u[i+1][j-1] +u[i+1][j+1] ) +
			     _lambda_tau *( u[i+2][j] + u[i-2][j] + u[i][j+2] + u[i][j-2] ) -
			     ( (float) beta[i][j]) * ((float) d[i][j]) );
	     }
	   else
	   if((i==0)&&(j==0))
	     {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      u[i][j] = u[i][j] - (w/T) *
			    ( T * u[i][j] -
			    _lambda_1_plus_3_tau * ( u[i+1][j] + u[i][j+1] ) +
                            _2_lambda_tau * u[i+1][j+1]
			    + _lambda_tau *
			    ( u[i+2][j] + u[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	     }
	   else
	   if((i==0)&&(j==cols-1))
	     {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      u[i][j] = u[i][j] - (w/T) *
			    ( T * u[i][j] -
			    _lambda_1_plus_3_tau * ( u[i+1][j] + u[i][j-1] ) +
			    _2_lambda_tau * u[i+1][j-1] +
			    _lambda_tau *( u[i+2][j] + u[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	   else
	   if((i==rows-1)&&(j==0))
	     {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      u[i][j] = u[i][j] - (w/T) *
			    ( T * u[i][j] -
			    _lambda_1_plus_3_tau * ( u[i-1][j] + u[i][j+1] ) +
			    _2_lambda_tau * u[i-1][j+1] +
			    _lambda_tau * ( u[i-2][j] + u[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	     }
	   else
	   if((i==rows-1)&&(j==cols-1))
	    {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      u[i][j] = u[i][j] - (w/T) *
			    ( T * u[i][j] -
			    _lambda_1_plus_3_tau *( u[i-1][j] + u[i][j-1] ) +
			    _2_lambda_tau * u[i-1][j-1] +
			    _lambda_tau *
			    ( u[i-2][j] + u[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((i==0)&&(j>1)&&(j<cols-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( u[i+1][j] + u[i][j-1] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i+1][j-1]  + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i][j-2] + u[i][j+2] + u[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==rows-1)&&(j>1)&&(j<cols-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( u[i][j-1] + u[i-1][j] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i-1][j+1] ) +
			    _lambda_tau * ( u[i][j-2] + u[i][j+2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
		    }
	  else
	  if((j==0)&&(i>1)&&(i<rows-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( u[i+1][j] + u[i-1][j] + u[i][j+1]) +
			    _2_lambda_tau * ( u[i-1][j+1]  + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i+2][j] + u[i][j+2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((j==cols-1)&&(i>1)&&(i<rows-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( u[i][j-1] + u[i-1][j] + u[i+1][j] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i+1][j-1] ) +
			    _lambda_tau * ( u[i+2][j] + u[i][j-2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==0)&&(j==1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i][j-1] - 
			    _lambda_1_plus_5_tau  *
			    ( u[i+1][j] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i+1][j-1]  + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i+2][j] + u[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==0)&&(j==cols-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i][j+1] -
			    _lambda_1_plus_5_tau  *
			    ( u[i][j-1] + u[i+1][j] ) +
			    _2_lambda_tau * ( u[i+1][j-1]  + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i][j-2] + u[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==1)&&(j==0))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i-1][j] -
			    _lambda_1_plus_5_tau  *
			    ( u[i+1][j] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i+1][j+1]  + u[i-1][j+1] ) +
			    _lambda_tau * ( u[i][j+2] + u[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==rows-2)&&(j==0))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i+1][j] -
			    _lambda_1_plus_5_tau  * ( u[i][j+1] + u[i-1][j] ) +
			    _2_lambda_tau * ( u[i+1][j+1]  + u[i-1][j+1] ) +
			    _lambda_tau * ( u[i][j+2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==rows-1)&&(j==1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i][j-1] -
			    _lambda_1_plus_5_tau  * ( u[i-1][j] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i-1][j+1] ) +
			    _lambda_tau * ( u[i][j+2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==rows-1)&&(j==cols-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i][j+1] -
			    _lambda_1_plus_5_tau  * ( u[i-1][j] + u[i][j-1] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i-1][j+1] ) +
			    _lambda_tau * ( u[i-2][j] + u[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==1)&&(j==cols-1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i-1][j] -
			    _lambda_1_plus_5_tau  * ( u[i][j-1] + u[i+1][j] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i+1][j-1] ) +
			    _lambda_tau * ( u[i][j-2] + u[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==rows-2)&&(j==cols-1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i+1][j] -
			    _lambda_1_plus_5_tau  * ( u[i-1][j] + u[i][j-1] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i+1][j-1] ) +
			    _lambda_tau * ( u[i][j-2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==1)&&(j==1))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			    ( T * u[i][j] -
			     _lambda_1_plus_5_tau * ( u[i][j-1] + u[i-1][j] ) -
			     _lambda_1_plus_7_tau * ( u[i+1][j] + u[i][j+1] ) +
			     _2_lambda_tau *
			     ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			     _lambda_tau * ( u[i][j+2] + u[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==rows-2)&&(j==1))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau * ( u[i][j-1]  + u[i+1][j] ) -
			    _lambda_1_plus_7_tau * ( u[i-1][j]  + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i-2][j] + u[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((i==1)&&(j==cols-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_7_tau * ( u[i][j-1] + u[i+1][j] ) -
			    _lambda_1_plus_5_tau * ( u[i-1][j] + u[i][j+1] ) + 
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i+2][j] + u[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
            }
	  else
	  if((i==rows-2)&&(j==cols-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_7_tau * ( u[i][j-1] + u[i-1][j] ) -
			    _lambda_1_plus_5_tau * ( u[i+1][j] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i-2][j] + u[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((i==1)&&(j>1)&&(j<cols-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_7_tau * ( u[i][j-1] + u[i+1][j] + u[i][j+1] ) -
			    _lambda_1_plus_5_tau * u[i-1][j]  +
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i][j+2] + u[i][j-2] + u[i+2][j]) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((i==rows-2)&&(j>1)&&(j<cols-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_7_tau * ( u[i][j-1] + u[i-1][j] + u[i][j+1] ) -
			    _lambda_1_plus_5_tau * u[i+1][j] +
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
		            _lambda_tau * ( u[i][j-2] + u[i][j+2] + u[i-2][j]) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((j==1)&&(i>1)&&(i<cols-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau * u[i][j-1] -
			    _lambda_1_plus_7_tau *(u[i-1][j] + u[i][j+1] +u[i+1][j] ) +
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i][j+2] + u[i-2][j] + u[i+2][j]) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((j==cols-2)&&(i>1)&&(i<cols-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_7_tau *
			    ( u[i][j-1] + u[i-1][j] + u[i+1][j] ) -
			    _lambda_1_plus_5_tau * u[i][j+1] +
			     _2_lambda_tau *
			    ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau *( u[i][j-2] + u[i-2][j] + u[i+2][j]) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
		    }
	  if(u[i][j]>255.0) u[i][j]=255.0 ;
	  if(u[i][j]<0.0) u[i][j]=0.0     ;
	  abs_error=fabs(u[i][j]-uc) ;
	  error+=abs_error;
	 } /* for */
	 error= error / ( (float) rows * (float) cols ) ;
         if((IN%16) == 0)
	 printf("\n error = %f",error);
	 IN++;
	}while(error>=d_error);
       printf("\n Iteration : %d",IN);
       printf("done.");

       for(i=0;i<rows;i++)
	  {
	   free(d[i]) ;
	  }

       free(      d) 	  ;

       image = alloc_2Darray(rows,cols,sizeof(unsigned char))	;

       for(i=0;i<rows;i++)
	  for(j=0;j<cols;j++)
             {
     	      if(beta[i][j] !=1 )
                {
    	         IN = rand()%2 					;
                 if(IN==0)
	           u[i][j]= u[i][j] + 0.5 			;
                 else
                   u[i][j]= u[i][j] - 0.5 		        ;
	         if(u[i][j]>255.0) u[i][j] = 255.0		;
		 if(u[i][j]<0.0) u[i][j] = 0.0			;
                }  
  	       image[i][j] = (unsigned char) u[i][j]		;              
	      }

       printf("\n writing to file ...");
       writehips(argv[2],image,rows,cols);
       printf("done.\n");
       for(i=0;i<rows;i++)
	  {
	   free(u    [i]) ;
	   free(beta[i]) ;
	   free(image[i]) ;
	  }
       free(      u)      ;
       free(image  )      ;
       free(beta   )	  ;
}
