/************************************************************************/
/* Optimal Sampling							*/
/************************************************************************/
#define Length  48
#define DLength 97
#define xnPoint	4

float g[DLength]		; /* Original function			*/
float Left[Length]		; /* Left Profile     			*/
float Right[Length]		; /* Right Profile    			*/
float f[DLength]		; /* Reconstructed function		*/
int   X[xnPoint]		; /* Unknown locations			*/
float c[xnPoint]		; /* Unknown coefficients		*/
float c_opt[xnPoint]		; /* Optimum coefficients		*/
float d[xnPoint]		; /* Data values at X			*/
float G[xnPoint][xnPoint]	; /* (symmetric)			*/
float M[Length][Length]		; /* Error Surface			*/
extern float Tau,Lambda		; /* Regularization Parameter		*/
Coordinate c_right,c_left						 ;

/************************************************************************/
/* Prototypes							        */
/************************************************************************/
void    FindEdgeParameters(Endpoint *,HipsImage ,HipsImage ,
       HipsImage ,HipsImage ,float ) ;
void    SetWidthAlongNormalAndTangentOfSpecifiedDirection(Coordinate ,int ,
        Direction ,float ,  float ,HipsImage , HipsImage ,HipsImage ,
	HipsImage ,int ,int );

void Solve(int)						;

/************************************************************************/
/* Implementations							*/
/************************************************************************/
void SetWidthAlongNormalAndTangentOfSpecifiedDirection(Coordinate c,
         int GL,
         Direction dir,
         float width_right,
         float width_left,
         HipsImage beta,
         HipsImage depth,
         HipsImage img,
         HipsImage eimg,
         int AvgRightContrast,
         int AvgLeftContrast)
{
    int        distance,dx,dy								;
    Direction  normal,pnormal	  							;
    Coordinate s,t									;
    bool       InsideDomain								;
    float      l,unit_distance								;

    normal  = dir 						 			;
    pnormal = RotateDirection(normal,_ANGLE_180)					;
    beta.UCHARMap[c.x][c.y] = _EDGE_255								;
    depth.UCHARMap[c.x][c.y] = GL								;
    printf("\n      Real (EDGE) - New %d %d",img.UCHARMap[c.x][c.y],GL) ;
    l = 0.0										;
    t = c										;
    unit_distance = 1.0	;
    do
      {
       dx  = unit_direction[normal][0]							;
       dy  = unit_direction[normal][1]							;
       s.x = t.x + dx                          						;
       s.y = t.y + dy                          						;
       InsideDomain = CheckCoordinate(s,img.Row,img.Column) 						;
       if(InsideDomain == _TRUE)
	 {
	  t   = s                                       				;
	  l  += unit_distance								;
	 }
      }
    while((l<width_right)&&(InsideDomain == _TRUE))					;
    if( eimg.UCHARMap[t.x][t.y] == _NON_EDGE)
      {
       if( ( GL + AvgRightContrast ) > 255 )
	    depth.UCHARMap[t.x][t.y] = 255							;
       else
       if( ( GL + AvgRightContrast ) < 0 )
	 depth.UCHARMap[t.x][t.y] = 0								;
       else
	   depth.UCHARMap[t.x][t.y] = (int) ( (float) GL +  (float) AvgRightContrast )					;
	beta.UCHARMap[t.x][t.y]  = _EDGE_255                                      		;
       }
    /* Other Normal Direction								*/
    l = 0.0										;
    t = c										;
    do
      {
       dx  = unit_direction[pnormal][0]							;
       dy  = unit_direction[pnormal][1]							;
       s.x = t.x + dx                          						;
       s.y = t.y + dy                          						;
       InsideDomain = CheckCoordinate(s,img.Row,img.Column) 						;
      if(InsideDomain == _TRUE)
	 {
	  t   = s                                       				;
	  l  += unit_distance								;
	 }
      }
    while((l<width_left)&&(InsideDomain == _TRUE))					;

    if( eimg.UCHARMap[t.x][t.y] == _NON_EDGE)
      {
       if(( GL + AvgLeftContrast ) > 256 )
	 depth.UCHARMap[t.x][t.y] = 255									;
       else
       if(( GL + AvgLeftContrast ) < 0 )
	 depth.UCHARMap[t.x][t.y] = 0									;
       else
	 depth.UCHARMap[t.x][t.y] =  (int) ( (float) GL + (float) AvgLeftContrast )	;
       beta.UCHARMap[t.x][t.y]  = _EDGE_255                                      				;
   }
}

void FindEdgeParameters(
     Endpoint *ptr,
     HipsImage img,
     HipsImage beta,
     HipsImage depth,
     HipsImage eimg,
     float T)
{
     Endpoint   *p									;
     ChainCode  *q,*qi									;
     int        dx,dy,i,j,k,sign,x,y,SL							;
     Direction  tangent,normal,pnormal,previous1,previous2            			;
     Coordinate r,s,t									;
     float      grt,grt_Previous,Dif_right,Dif_left,AvgRightContrast,AvgLeftContrast 	;
     float      unit_distance,l,w_right,w_left						;
     bool       InsideDomain,Change,bQuit,Current,bFull = _FALSE			;
     int	m,n,X1min,X2min,left,right						;
     float	Mmin = 0.0								;
   
     p = ptr										;
     if(p!=NULL)
       do {
	   q = p->chain									;
	   r = q->point									;
	   SL = GetSegmentLength(q) - 1 						;
	   /* Block Length = 0								*/
	   if(q!=NULL)
	     do {
	     qi = q									;
	     k = 0									;
	     /**************************************************************************/
	     /* 1st pass : to extract Edge Profiles to be modeled              	       */
	     /**************************************************************************/
	     do
	       {
		if(qi==p->chain)
		  previous1 = qi->dir						;
		tangent = qi->dir 						;
               if( (tangent%2) == 1) {
                  if( ( (tangent == 3) || (tangent == 7) ) && ((previous1 == 2)||(previous1 ==6)||(previous1 ==1)||(previous1 ==5)))
                    normal = GetNormalDirection((tangent-1)%8)                          ;
                  else
                    if( ( (tangent == 1) || (tangent == 5) ) && ((previous1 == 0)||(previous1 ==4)||(previous1==1)||(previous1==3)))
                      normal = GetNormalDirection((tangent-1)%8)                                ;
                    else
                      normal = GetNormalDirection((tangent+1)%8)                                ; }
                else
	        normal = GetNormalDirection(tangent)				;

		previous1 = tangent	;
		pnormal = RotateDirection(normal,_ANGLE_180)			;
		/* Normal Direction					       */
		l = 0.0								;
		t = r								;
		k++								;

		left = 0						;
		unit_distance = 1.0					;

		Left[left]    = (float)img.UCHARMap[t.x][t.y]			;
		dx  = unit_direction[normal][0]				;
		dy  = unit_direction[normal][1]				;
		s.x = t.x + dx						;
		s.y = t.y + dy					       	;
		InsideDomain = CheckCoordinate(s,img.Row,img.Column)			;
		if( InsideDomain == _TRUE)
		  {
		   left++						;
		   Left[left]    = (float)img.UCHARMap[t.x][t.y]			;
		   grt = (float)img.UCHARMap[s.x][s.y] - (float)img.UCHARMap[t.x][t.y]   	;
		   if(grt<0.0) {
		      grt = 0.0 - grt;
		      Change = _FALSE;}
		   else Change = _TRUE	;
		      t   = s						       	;
		      l  += unit_distance					;
		  }

                  grt_Previous = grt 						;
                  bQuit = _TRUE	;
		do
		  {
		   dx  = unit_direction[normal][0]				;
		   dy  = unit_direction[normal][1]				;
		   s.x = t.x + dx						;
		   s.y = t.y + dy					       	;
		   InsideDomain = CheckCoordinate(s,img.Row,img.Column)		;
		   if( InsideDomain == _TRUE)
		     {
		      grt = (float)img.UCHARMap[s.x][s.y] - (float)img.UCHARMap[t.x][t.y]   	;
		      t   = s						       	;
		      l  += unit_distance					;
		      if(grt<0.0) {
			grt = 0.0 - grt						;
			Current = _FALSE ; }
		      else Current = _TRUE					;
		      if( Current == Change ) {
			 left++							;
			 Left[left]    = (float)img.UCHARMap[s.x][s.y]			;
			t  = s						       	;
			l += unit_distance					;
		       }
		    }
                    if((grt<=T) && (grt_Previous<=T))
                      bQuit = _FALSE ;
                    grt_Previous = grt ;
		  }
		while(((grt>=T)||bQuit)&&(l<(float)Length)&&(InsideDomain == _TRUE)&&(Current == Change));

		/***********************************************************************/
		/* Other normal direction 				       	       */
		/***********************************************************************/
		l = 0.0									;
		t = r									;
		right = 0						;
		dx  = unit_direction[pnormal][0]			;
		dy  = unit_direction[pnormal][1]			;
		s.x = t.x + dx						;
		s.y = t.y + dy					       	;
		InsideDomain = CheckCoordinate(s,img.Row,img.Column)	;
		if( InsideDomain == _TRUE)
		  {
		   Right[right]   = (float)img.UCHARMap[s.x][s.y]			;
		   right++						;
		   grt = (float)img.UCHARMap[s.x][s.y] - (float)img.UCHARMap[t.x][t.y]   	;
		   if(grt<0.0) {
		      grt = 0.0 - grt;
		      Change = _FALSE;}
		   else Change = _TRUE	;
		      t   = s						       	;
		      l  += unit_distance					;
		     }
                bQuit = _TRUE	;
                grt_Previous = grt 						;

		do
		 {
		  dx  = unit_direction[pnormal][0]				;
		  dy  = unit_direction[pnormal][1]				;
		  s.x = t.x + dx						;
		  s.y = t.y + dy					       	;
		  InsideDomain = CheckCoordinate(s,img.Row,img.Column)		;
		  if( InsideDomain == _TRUE)
		    {
		     grt = (float)img.UCHARMap[s.x][s.y] - (float)img.UCHARMap[t.x][t.y]   	;
		     if(grt<0.0) {
		       grt = 0.0 - grt						;
		       Current = _FALSE ; }
		     else Current = _TRUE					;
		     if(Current == Change){
		       Right[right]   = (float)img.UCHARMap[s.x][s.y]		;
		       right++						;
		       t = s						       	;
		       l+= unit_distance					;
		      }
		    }
                    if((grt<=T) && (grt_Previous<=T))
                      bQuit = _FALSE ;
                    grt_Previous = grt ;
		 }
	       while(((grt>=T)||bQuit)&&(l<(float)Length)&&(InsideDomain == _TRUE)&&(Current==Change))	;

     for(i=0;i<=left;i++)
	g[i] = Left[left-i]		;
     for(i=left+1;i<=left+right;i++)
	g[i] = Right[i-left-1]		;
     /* Finding Optimum Sampling*/
     X[1] = left 	;							;

     for(m=0;m<left;m++){
	for(n=left+1;n<=left+right;n++){

     X[0] = m ; X[2] = n 				;
     M[left-X[0]-1][X[2]-left-1] = 0.0			;

     for(i=0;i<xnPoint-1;i++)
	for(j=0;j<xnPoint-1;j++)
	   G[i][j] = abs((float)X[i] - (float)X[j])	;


     for(i=0;i<xnPoint-1;i++)
	G[i][i] = Lambda 				;

     for(i=0;i<xnPoint-1;i++) {
	G[i][xnPoint-1] = 1.0				;
	d[i]            = g[X[i]]			; }

     for(i=0;i<xnPoint-1;i++)
	G[xnPoint-1][i] = 1.0				;

     for(i=0;i<xnPoint-1;i++)
	G[i][xnPoint-1] = 1.0				;

     G[xnPoint-1][xnPoint-1] = 0.0			;
     d[xnPoint-1]  	     = 0.0			;

     Solve(xnPoint)					;

     for(i=0;i<=left+right;i++){
	   f[i] = c[0] *  abs((float)i - (float)X[0]) 	;
	   f[i] += c[1] * abs((float)i - (float)X[1]) ;
	   f[i] += c[2] * abs((float)i - (float)X[2]) + c[3]	;

	if(f[i]>255.0) f[i]=255.0	;
	if(f[i]<0.0) f[i]=0.0	;

	M[left-X[0]-1][X[2]-left-1] += (f[i]-g[i]) * (f[i]-g[i]) ; }
	M[left-X[0]-1][X[2]-left-1] /= (float) (right+left+1)			;

     if( (m==0) && (n==left+1) ){
     for(i=0;i<xnPoint;i++)
	c_opt[i] = c[i]				;
	Mmin = M[left-X[0]-1][X[2]-left-1]	;
	X1min = X[0]				;
	X2min = X[2]            		; }

     if(M[left-X[0]-1][X[2]-left-1]<=Mmin){
     for(i=0;i<xnPoint;i++)
	c_opt[i] = c[i]			;
     Mmin = M[left-X[0]-1][X[2]-left-1]	;
     X1min = X[0]			;
     X2min = X[2]               	; }}


     for(i=0;i<=left+right;i++){
	   f[i] = c_opt[0] *  abs((float)i - (float)X1min) 		;
	   f[i] += c_opt[1] *  abs((float)i - (float)X[1])		;
	   f[i] += c_opt[2] *  abs((float)i - (float)X2min) + c_opt[3]	;
	if(f[i]>255.0) f[i]=255.0	;
	if(f[i]<0.0) f[i]=0.0		;

	if(i<left) {
	   dx  = unit_direction[normal][0]			;
	   dy  = unit_direction[normal][1]			;
	   s.x = r.x + (left-i) * dx				;
	   s.y = r.y + (left-i) * dy			       	;
	   depth.UCHARMap[s.x][s.y] = (unsigned char)f[i]		;
	   beta.UCHARMap[s.x][s.y] = _EDGE_255				; }
	   else {
	   dx  = unit_direction[pnormal][0]			;
	   dy  = unit_direction[pnormal][1]			;
	   s.x = r.x + (i-left) * dx				;
	   s.y = r.y + (i-left) * dy			       	;
	   depth.UCHARMap[s.x][s.y] = (unsigned char) f[i]		;
	   beta.UCHARMap[s.x][s.y] = _EDGE_255				; }
	}}
	       qi = qi->next 								;
	       if(qi!=NULL) r = qi->point						;
	      }
	    while(qi!=NULL)				;
	   q = qi											;
	  }while(q!=NULL)								;
	  p = p->next											;
	 }
       while(p!=NULL)			;								;
}

void Solve(int N)
{
     int i,j,k		;
     float pivot,a	;

     for(i=0;i<N-1;i++)
	for(j=i+1;j<N;j++){
          if(G[i][i] != 0.0)
	    pivot = -G[j][i]/G[i][i]		;
          else
	    pivot = 0.0		                ;
	  G[j][i] = 0.0				;
	    for(k=i+1;k<N;k++)
	       G[j][k] += pivot * G[i][k]	;
	    d[j] += pivot * d[i] 	  	; }

    if(G[N-1][N-1] != 0.0)
      c[N-1] = d[N-1] / G[N-1][N-1]		;
    else
      c[N-1] = 0.0                              ;
     for(k=N-2;k>=0;k--) {
        c[k] = d[k]                             ;
        for(i=N-1;i>k;i--)
           c[k] -= G[k][i] * c[i]               ;

       if(G[k][k] != 0.0)
           c[k] /= G[k][k]                         ;
       else
           c[k] = 0.0                           ;         }
}
