/***************************************************************************/
/* Surface Reconstruction By Hybrid Model				   */
/* 			     ( Membrane & Thin-plate )		    	   */
/***************************************************************************/
#include "Image.h"

class SOR:public image{
      public:
            SOR()                                               ;
           ~SOR()                                               ;
      void ReconstructTheSurface(void)                          ;
      void SetParameters(float,float,float,float,bool)          ;
      private:
      unsigned char     **depth,**beta  	                ;
      float		**u      				;
      float tau,lambda,error,d_error,abs_error,w,T,uc 	        ;
      bool boolInterpolate                                      ;
}

void SOR::SOR(void){
     boolInterpolate = false    ;
}

void SOR::~SOR(void){
     int i ;
     for(i=0;i<Row;i++)
        delete u[i] ;
     delete u ;
     for(i=0;i<Row;i++)
        delete Y[i] ;
     delete Y ;
}

void SOR::ReconstructTheSurface(void)
{
    int   	  IN=0 						;
    float average,_lambda_3_plus_8_tau				;
    float _4_lambda_1_plus_4_tau,_lambda_1_plus_7_tau		;
    float _2_lambda_tau,_lambda_tau,_2_lambda_1_plus_tau	;
    float _lambda_1_plus_3_tau,_lambda_1_plus_5_tau		;
    float _lambda_4_plus_14_tau,_lambda_4_plus_15_tau		;
    float _lambda_3_plus_6_tau,_lambda_3_plus_7_tau		;
    float A,B							;


    u = new float* (Row)        ;
    for(i=0;i++;i<Row)
       u[i] = new float (Column)        ;

    y = new unsigned char* (Row)        ;
    for(i=0;i++;i<Row)
       u[i] = new unsigned char (Column)        ;

    for(i=0;i<Row;i++)
	for(j=0;j<Column;j++)
	   {
	    if(beta[i][j]>0)
	      beta[i][j] = 1 ;
            if(beta[i][j] == 0)
             u[i][j] = 128.0  ;
            else
       	     u[i][j]=(float) depth[i][j] ;
	   }

    _4_lambda_1_plus_4_tau = 4.0 * lambda * ( 1.0 + 4.0 * tau )   	;
    _lambda_3_plus_7_tau   = lambda * (3.0 + 7.0 * tau ) 	  	;
    _lambda_1_plus_7_tau   = lambda * (1.0 + 7.0 * tau ) 	  	;
    _lambda_1_plus_5_tau   = lambda * (1.0 + 5.0 * tau ) 	  	;
    _lambda_1_plus_3_tau   = lambda * (1.0 + 3.0 * tau ) 	  	;
    _lambda_4_plus_14_tau  = lambda * (4.0 +14.0 * tau ) 	  	;
    _lambda_3_plus_8_tau   = lambda * (3.0 + 8.0 * tau ) 	  	;
    _lambda_3_plus_6_tau   = lambda * (3.0 + 6.0 * tau ) 	  	;
    _lambda_4_plus_15_tau  = lambda * (4.0 +15.0 * tau ) 	  	;
    _2_lambda_1_plus_tau   = 2.0 * lambda * ( 1.0 + tau ) 	  	;
    _lambda_tau	           = lambda * tau			       	;
    _2_lambda_tau	   = 2.0* lambda * tau		  		;
    A = lambda * (1-tau)						;
    B = lambda * tau							;

    do
      {
       error=0.0 ;
       for(i=0;i<Row;i++)
       for(j=0;j<Column;j++)
	  {
	   uc=u[i][j] ;
	   if(( (boolInterpolate==true)&&(betha[i][j]==0)) || (boolInterpolate==false))
	   if((i>1)&&(i<Row-2)&&(j>1)&&(j<Column-2))
	     {
              T = beta[i][j] + 4.0 * A + 16.0 * B  	;
              u[i][j] = u[i][j] - (w/T) * ( -((float)beta[i][j] * (float)d[i][j])
					    + T * u[i][j]
					    - ( A + B * (3.0*M[i-1][j]+2.0*M[i][j]+M[i-1][j-1])) * u[i-1][j]
					    - ( A + B * (2.0*M[i+1][j]+3.0*M[i][j]+M[i][j-1]  )) * u[i+1][j]
					    - ( A + B * (3.0*M[i][j-1]+2.0*M[i][j]+M[i-1][j-1])) * u[i][j-1]
					    - ( A + B * (3.0*M[i][j]+2.0*M[i][j+1]+M[i-1][j]  )) * u[i][j+1]
					    + B * M[i-1][j-1] * u[i-1][j-1]
					    + B * M[i][j]     * u[i+1][j+1]
					    + B * M[i-1][j]   * u[i-1][j+1]
					    + B * M[i][j-1]   * u[i+1][j-1]
					    + B * M[i][j-1]   * u[i][j-2]
					    + B * M[i+1][j]   * u[i+1][j]
					    + B * M[i-1][j]   * u[i-2][j]
					    + B * M[i][j+1]   * u[i][j+2]) ;
	     }
	   else
	   if((i==0)&&(j==0))
	     {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      u[i][j] = u[i][j] - (w/T) *
			    ( T * u[i][j] -
			    _lambda_1_plus_3_tau * ( u[i+1][j] + u[i][j+1] ) +
                            _2_lambda_tau * u[i+1][j+1]
			    + _lambda_tau *
			    ( u[i+2][j] + u[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	     }
	   else
	   if((i==0)&&(j==Column-1))
	     {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      u[i][j] = u[i][j] - (w/T) *
			    ( T * u[i][j] -
			    _lambda_1_plus_3_tau * ( u[i+1][j] + u[i][j-1] ) +
			    _2_lambda_tau * u[i+1][j-1] +
			    _lambda_tau *( u[i+2][j] + u[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	   else
	   if((i==Row-1)&&(j==0))
	     {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      u[i][j] = u[i][j] - (w/T) *
			    ( T * u[i][j] -
			    _lambda_1_plus_3_tau * ( u[i-1][j] + u[i][j+1] ) +
			    _2_lambda_tau * u[i-1][j+1] +
			    _lambda_tau * ( u[i-2][j] + u[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	     }
	   else
	   if((i==Row-1)&&(j==Column-1))
	    {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      u[i][j] = u[i][j] - (w/T) *
			    ( T * u[i][j] -
			    _lambda_1_plus_3_tau *( u[i-1][j] + u[i][j-1] ) +
			    _2_lambda_tau * u[i-1][j-1] +
			    _lambda_tau *
			    ( u[i-2][j] + u[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((i==0)&&(j>1)&&(j<Column-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( u[i+1][j] + u[i][j-1] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i+1][j-1]  + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i][j-2] + u[i][j+2] + u[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==Row-1)&&(j>1)&&(j<Column-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( u[i][j-1] + u[i-1][j] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i-1][j+1] ) +
			    _lambda_tau * ( u[i][j-2] + u[i][j+2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
		    }
	  else
	  if((j==0)&&(i>1)&&(i<Row-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( u[i+1][j] + u[i-1][j] + u[i][j+1]) +
			    _2_lambda_tau * ( u[i-1][j+1]  + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i+2][j] + u[i][j+2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((j==Column-1)&&(i>1)&&(i<Row-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( u[i][j-1] + u[i-1][j] + u[i+1][j] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i+1][j-1] ) +
			    _lambda_tau * ( u[i+2][j] + u[i][j-2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==0)&&(j==1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i][j-1] - 
			    _lambda_1_plus_5_tau  *
			    ( u[i+1][j] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i+1][j-1]  + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i+2][j] + u[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==0)&&(j==Column-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i][j+1] -
			    _lambda_1_plus_5_tau  *
			    ( u[i][j-1] + u[i+1][j] ) +
			    _2_lambda_tau * ( u[i+1][j-1]  + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i][j-2] + u[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==1)&&(j==0))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i-1][j] -
			    _lambda_1_plus_5_tau  *
			    ( u[i+1][j] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i+1][j+1]  + u[i-1][j+1] ) +
			    _lambda_tau * ( u[i][j+2] + u[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==Row-2)&&(j==0))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i+1][j] -
			    _lambda_1_plus_5_tau  * ( u[i][j+1] + u[i-1][j] ) +
			    _2_lambda_tau * ( u[i+1][j+1]  + u[i-1][j+1] ) +
			    _lambda_tau * ( u[i][j+2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==Row-1)&&(j==1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i][j-1] -
			    _lambda_1_plus_5_tau  * ( u[i-1][j] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i-1][j+1] ) +
			    _lambda_tau * ( u[i][j+2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==Row-1)&&(j==Column-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i][j+1] -
			    _lambda_1_plus_5_tau  * ( u[i-1][j] + u[i][j-1] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i-1][j+1] ) +
			    _lambda_tau * ( u[i-2][j] + u[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==1)&&(j==Column-1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i-1][j] -
			    _lambda_1_plus_5_tau  * ( u[i][j-1] + u[i+1][j] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i+1][j-1] ) +
			    _lambda_tau * ( u[i][j-2] + u[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==Row-2)&&(j==Column-1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
                            _lambda_1_plus_3_tau  * u[i+1][j] -
			    _lambda_1_plus_5_tau  * ( u[i-1][j] + u[i][j-1] ) +
			    _2_lambda_tau * ( u[i-1][j-1]  + u[i+1][j-1] ) +
			    _lambda_tau * ( u[i][j-2] + u[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==1)&&(j==1))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			    ( T * u[i][j] -
			     _lambda_1_plus_5_tau * ( u[i][j-1] + u[i-1][j] ) -
			     _lambda_1_plus_7_tau * ( u[i+1][j] + u[i][j+1] ) +
			     _2_lambda_tau *
			     ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			     _lambda_tau * ( u[i][j+2] + u[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]));
	    }
	  else
	  if((i==Row-2)&&(j==1))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau * ( u[i][j-1]  + u[i+1][j] ) -
			    _lambda_1_plus_7_tau * ( u[i-1][j]  + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i-2][j] + u[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((i==1)&&(j==Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_7_tau * ( u[i][j-1] + u[i+1][j] ) -
			    _lambda_1_plus_5_tau * ( u[i-1][j] + u[i][j+1] ) + 
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i+2][j] + u[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
            }
	  else
	  if((i==Row-2)&&(j==Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_7_tau * ( u[i][j-1] + u[i-1][j] ) -
			    _lambda_1_plus_5_tau * ( u[i+1][j] + u[i][j+1] ) +
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i-2][j] + u[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((i==1)&&(j>1)&&(j<Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_7_tau * ( u[i][j-1] + u[i+1][j] + u[i][j+1] ) -
			    _lambda_1_plus_5_tau * u[i-1][j]  +
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i][j+2] + u[i][j-2] + u[i+2][j]) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((i==Row-2)&&(j>1)&&(j<Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_7_tau * ( u[i][j-1] + u[i-1][j] + u[i][j+1] ) -
			    _lambda_1_plus_5_tau * u[i+1][j] +
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
		            _lambda_tau * ( u[i][j-2] + u[i][j+2] + u[i-2][j]) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((j==1)&&(i>1)&&(i<Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_5_tau * u[i][j-1] -
			    _lambda_1_plus_7_tau *(u[i-1][j] + u[i][j+1] +u[i+1][j] ) +
			    _2_lambda_tau * ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau * ( u[i][j+2] + u[i-2][j] + u[i+2][j]) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
	    }
	  else
	  if((j==Column-2)&&(i>1)&&(i<Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     u[i][j]=u[i][j] - (w/T) *
			   ( T * u[i][j] -
			    _lambda_1_plus_7_tau *
			    ( u[i][j-1] + u[i-1][j] + u[i+1][j] ) -
			    _lambda_1_plus_5_tau * u[i][j+1] +
			     _2_lambda_tau *
			    ( u[i-1][j-1] + u[i-1][j+1] + u[i+1][j-1] + u[i+1][j+1] ) +
			    _lambda_tau *( u[i][j-2] + u[i-2][j] + u[i+2][j]) -
			    ((float) beta[i][j]) * ((float) d[i][j]) );
		    }
	  if(u[i][j]>255.0) u[i][j]=255.0 ;
	  if(u[i][j]<0.0) u[i][j]=0.0     ;
	  abs_error=fabs(u[i][j]-uc) ;
	  error+=abs_error;
	 } /* for */
	 error= error / ( (float) Row * (float) Column ) ;
	 IN++;
	}while(error>=d_error);

       for(i=0;i<Row;i++)
	  for(j=0;j<Column;j++)
             {
     	      if(beta[i][j] !=1 )
                {
    	         IN = rand()%2 					;
                 if(IN==0)
	           u[i][j]= u[i][j] + 0.5 			;
                 else
                   u[i][j]= u[i][j] - 0.5 		        ;
	         if(u[i][j]>255.0) u[i][j] = 255.0		;
		 if(u[i][j]<0.0) u[i][j] = 0.0			;
                }
	      }
}
