/***************************************************************************/
/* 			     		    	                           */
/*  Surface Reconstruction By Hybrid Model				   */
/* 			     ( Membrane + Thin-plate )		    	   */
/* 			     		    	                           */
/***************************************************************************/
/* 			     		    	                           */
/*  SOR Class Library          	                                           */
/* 			     		    	                           */
/*     SOR              	     	    	                           */
/*      |              	     		    	                           */
/*    image              	     	    	                           */
/* 			     		    	                           */
/***************************************************************************/
/* 			     		    	                           */
/*  Last modified by    		    	                           */
/*                   Binnur KURT            	                           */
/*  When        	     		    	                           */
/* 		     August 19th 1999    	                           */
/* 			     		    	                           */
/***************************************************************************/
#include "image.h"

class SOR:public Image{
      public:
            SOR()                                               ;
           ~SOR()                                               ;
      void ReconstructTheSurface(void)                          ;
      void SetParameters(float,float,float,float,bool)          ;
      void AllocateMemoryForModelMatrices(void)                 ;
      void SaveDepthAndBetaToTImage(TImage *,TImage *)          ;
      unsigned char     **depth,**beta  	                ;
      int  IterationNumber                                      ;

      private:
      float		**surface      				;
      float tau,lambda,error,d_error,abs_error,w,T,uc 	        ;
      bool boolInterpolate                                      ;
      bool boolIsReconstructed                                  ;
} ;

      SOR::SOR(){
      boolInterpolate = false                                   ;
      boolIsReconstructed = false                               ;
      IterationNumber = 0                                       ;
}

SOR::~SOR(){
    int i ;
    if (boolIsReconstructed == true){
       delete[] surface[0];
       delete[] surface;
    }
}

void SOR::SetParameters(float parLambda, float parTau ,float parW,
                        float parDesiredError, bool parIsInterpolation){
     lambda  = parLambda        ;
     tau     = parTau           ;
     w       = parW             ;
     d_error = parDesiredError  ;
     boolInterpolate = parIsInterpolation       ;
}

void SOR::ReconstructTheSurface(void)
{
    int   i,j                                                   ;
    float _lambda_3_plus_8_tau				        ;
    float _lambda_1_plus_7_tau,_4_lambda_1_plus_4_tau		;
    float _2_lambda_tau,_lambda_tau,_2_lambda_1_plus_tau	;
    float _lambda_1_plus_3_tau,_lambda_1_plus_5_tau		;
    float _lambda_4_plus_14_tau,_lambda_4_plus_15_tau		;
    float _lambda_3_plus_7_tau		                        ;
    float _lambda = 0.01                                         ;
    
    boolIsReconstructed = true                                  ;

    surface = new float*[Row];
    surface[0] = new float[Row*Column];
    float *p= surface[0]+Column;
    for(i=1;i<Row;i++,p+=Column)
       surface[i] = p;
    unsigned char *pBeta=beta[0];
    unsigned char *pDepth=depth[0];
    float *pSurface=surface[0];
    for (i=0;i<Row;i++){
	for (j=0;j<Column;j++,pDepth++,pBeta++,pSurface++){
	     if (*pBeta>0)  *pBeta=1;
             if (*pBeta==0) *pSurface= 128.0;
             else           *pSurface=(float) *pDepth;
	   }
    }

    do
      {
       error=0.0 ;
       for(i=0;i<Row;i++)
       for(j=0;j<Column;j++)
	  {
	   uc=surface[i][j] ;
//	   if(( (boolInterpolate==true)&&(beta[i][j]==0)) || (boolInterpolate==false))
    if(beta[i][j]==0){
    _4_lambda_1_plus_4_tau = 4.0 * lambda * (1.0 + 4.0 * tau ) 	  	;
    _lambda_3_plus_7_tau   = lambda * (3.0 + 7.0 * tau ) 	  	;
    _lambda_1_plus_7_tau   = lambda * (1.0 + 7.0 * tau ) 	  	;
    _lambda_1_plus_5_tau   = lambda * (1.0 + 5.0 * tau ) 	  	;
    _lambda_1_plus_3_tau   = lambda * (1.0 + 3.0 * tau ) 	  	;
    _lambda_4_plus_14_tau  = lambda * (4.0 +14.0 * tau ) 	  	;
    _lambda_3_plus_8_tau   = lambda * (3.0 + 8.0 * tau ) 	  	;
    _lambda_4_plus_15_tau  = lambda * (4.0 +15.0 * tau ) 	  	;
    _2_lambda_1_plus_tau   = 2.0 * lambda * ( 1.0 + tau ) 	  	;
    _lambda_tau	           = lambda * tau			       	;
    _2_lambda_tau	   = 2.0* lambda * tau		  		;
    }
    else{
    _4_lambda_1_plus_4_tau = 4.0 * _lambda * (1.0 + 4.0 * tau ) 	  	;
    _lambda_3_plus_7_tau   = _lambda * (3.0 + 7.0 * tau ) 	  	;
    _lambda_1_plus_7_tau   = _lambda * (1.0 + 7.0 * tau ) 	  	;
    _lambda_1_plus_5_tau   = _lambda * (1.0 + 5.0 * tau ) 	  	;
    _lambda_1_plus_3_tau   = _lambda * (1.0 + 3.0 * tau ) 	  	;
    _lambda_4_plus_14_tau  = _lambda * (4.0 +14.0 * tau ) 	  	;
    _lambda_3_plus_8_tau   = _lambda * (3.0 + 8.0 * tau ) 	  	;
    _lambda_4_plus_15_tau  = _lambda * (4.0 +15.0 * tau ) 	  	;
    _2_lambda_1_plus_tau   = 2.0 * _lambda * ( 1.0 + tau ) 	  	;
    _lambda_tau	           = _lambda * tau			       	;
    _2_lambda_tau	   = 2.0* _lambda * tau		  		;
    }
	   if((i>1)&&(i<Row-2)&&(j>1)&&(j<Column-2))
	     {
              T = beta[i][j] + _4_lambda_1_plus_4_tau ;
              surface[i][j] = surface[i][j] - (w/T) *
	   	    ( T * surface[i][j] -
	   	     _lambda_1_plus_7_tau  * ( surface[i-1][j] + surface[i][j-1] + surface[i+1][j] + surface[i][j+1] ) +
	   	     _2_lambda_tau *( surface[i-1][j-1] +surface[i-1][j+1] +surface[i+1][j-1] +surface[i+1][j+1] ) +
	   	     _lambda_tau *( surface[i+2][j] + surface[i-2][j] + surface[i][j+2] + surface[i][j-2] ) -
	   	     ( (float) beta[i][j]) * ((float) depth[i][j]) );
	     }
	   else
	   if((i==0)&&(j==0))
	     {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      surface[i][j] = surface[i][j] - (w/T) *
			    ( T * surface[i][j] -
			    _lambda_1_plus_3_tau * ( surface[i+1][j] + surface[i][j+1] ) +
                            _2_lambda_tau * surface[i+1][j+1]
			    + _lambda_tau *
			    ( surface[i+2][j] + surface[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
	     }
	   else
	   if((i==0)&&(j==Column-1))
	     {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      surface[i][j] = surface[i][j] - (w/T) *
			    ( T * surface[i][j] -
			    _lambda_1_plus_3_tau * ( surface[i+1][j] + surface[i][j-1] ) +
			    _2_lambda_tau * surface[i+1][j-1] +
			    _lambda_tau *( surface[i+2][j] + surface[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
	    }
	   else
	   if((i==Row-1)&&(j==0))
	     {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      surface[i][j] = surface[i][j] - (w/T) *
			    ( T * surface[i][j] -
			    _lambda_1_plus_3_tau * ( surface[i-1][j] + surface[i][j+1] ) +
			    _2_lambda_tau * surface[i-1][j+1] +
			    _lambda_tau * ( surface[i-2][j] + surface[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
	     }
	   else
	   if((i==Row-1)&&(j==Column-1))
	    {
	      T=(float)beta[i][j]+ _2_lambda_1_plus_tau  ;
	      surface[i][j] = surface[i][j] - (w/T) *
			    ( T * surface[i][j] -
			    _lambda_1_plus_3_tau *( surface[i-1][j] + surface[i][j-1] ) +
			    _2_lambda_tau * surface[i-1][j-1] +
			    _lambda_tau *
			    ( surface[i-2][j] + surface[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
	    }
	  else
	  if((i==0)&&(j>1)&&(j<Column-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( surface[i+1][j] + surface[i][j-1] + surface[i][j+1] ) +
			    _2_lambda_tau * ( surface[i+1][j-1]  + surface[i+1][j+1] ) +
			    _lambda_tau * ( surface[i][j-2] + surface[i][j+2] + surface[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==Row-1)&&(j>1)&&(j<Column-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( surface[i][j-1] + surface[i-1][j] + surface[i][j+1] ) +
			    _2_lambda_tau * ( surface[i-1][j-1]  + surface[i-1][j+1] ) +
			    _lambda_tau * ( surface[i][j-2] + surface[i][j+2] + surface[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
		    }
	  else
	  if((j==0)&&(i>1)&&(i<Row-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( surface[i+1][j] + surface[i-1][j] + surface[i][j+1]) +
			    _2_lambda_tau * ( surface[i-1][j+1]  + surface[i+1][j+1] ) +
			    _lambda_tau * ( surface[i+2][j] + surface[i][j+2] + surface[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((j==Column-1)&&(i>1)&&(i<Row-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_8_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_5_tau  *
			    ( surface[i][j-1] + surface[i-1][j] + surface[i+1][j] ) +
			    _2_lambda_tau * ( surface[i-1][j-1]  + surface[i+1][j-1] ) +
			    _lambda_tau * ( surface[i+2][j] + surface[i][j-2] + surface[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==0)&&(j==1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
                            _lambda_1_plus_3_tau  * surface[i][j-1] -
			    _lambda_1_plus_5_tau  *
			    ( surface[i+1][j] + surface[i][j+1] ) +
			    _2_lambda_tau * ( surface[i+1][j-1]  + surface[i+1][j+1] ) +
			    _lambda_tau * ( surface[i+2][j] + surface[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==0)&&(j==Column-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
                            _lambda_1_plus_3_tau  * surface[i][j+1] -
			    _lambda_1_plus_5_tau  *
			    ( surface[i][j-1] + surface[i+1][j] ) +
			    _2_lambda_tau * ( surface[i+1][j-1]  + surface[i+1][j+1] ) +
			    _lambda_tau * ( surface[i][j-2] + surface[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==1)&&(j==0))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
                            _lambda_1_plus_3_tau  * surface[i-1][j] -
			    _lambda_1_plus_5_tau  *
			    ( surface[i+1][j] + surface[i][j+1] ) +
			    _2_lambda_tau * ( surface[i+1][j+1]  + surface[i-1][j+1] ) +
			    _lambda_tau * ( surface[i][j+2] + surface[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==Row-2)&&(j==0))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
                            _lambda_1_plus_3_tau  * surface[i+1][j] -
			    _lambda_1_plus_5_tau  * ( surface[i][j+1] + surface[i-1][j] ) +
			    _2_lambda_tau * ( surface[i+1][j+1]  + surface[i-1][j+1] ) +
			    _lambda_tau * ( surface[i][j+2] + surface[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==Row-1)&&(j==1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
                            _lambda_1_plus_3_tau  * surface[i][j-1] -
			    _lambda_1_plus_5_tau  * ( surface[i-1][j] + surface[i][j+1] ) +
			    _2_lambda_tau * ( surface[i-1][j-1]  + surface[i-1][j+1] ) +
			    _lambda_tau * ( surface[i][j+2] + surface[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==Row-1)&&(j==Column-2))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
                            _lambda_1_plus_3_tau  * surface[i][j+1] -
			    _lambda_1_plus_5_tau  * ( surface[i-1][j] + surface[i][j-1] ) +
			    _2_lambda_tau * ( surface[i-1][j-1]  + surface[i-1][j+1] ) +
			    _lambda_tau * ( surface[i-2][j] + surface[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==1)&&(j==Column-1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
                            _lambda_1_plus_3_tau  * surface[i-1][j] -
			    _lambda_1_plus_5_tau  * ( surface[i][j-1] + surface[i+1][j] ) +
			    _2_lambda_tau * ( surface[i-1][j-1]  + surface[i+1][j-1] ) +
			    _lambda_tau * ( surface[i][j-2] + surface[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==Row-2)&&(j==Column-1))
	    {
	     T=(float)beta[i][j]+_lambda_3_plus_7_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
                            _lambda_1_plus_3_tau  * surface[i+1][j] -
			    _lambda_1_plus_5_tau  * ( surface[i-1][j] + surface[i][j-1] ) +
			    _2_lambda_tau * ( surface[i-1][j-1]  + surface[i+1][j-1] ) +
			    _lambda_tau * ( surface[i][j-2] + surface[i-2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==1)&&(j==1))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			    ( T * surface[i][j] -
			     _lambda_1_plus_5_tau * ( surface[i][j-1] + surface[i-1][j] ) -
			     _lambda_1_plus_7_tau * ( surface[i+1][j] + surface[i][j+1] ) +
			     _2_lambda_tau *
			     ( surface[i-1][j-1] + surface[i-1][j+1] + surface[i+1][j-1] + surface[i+1][j+1] ) +
			     _lambda_tau * ( surface[i][j+2] + surface[i+2][j] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]));
	    }
	  else
	  if((i==Row-2)&&(j==1))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_5_tau * ( surface[i][j-1]  + surface[i+1][j] ) -
			    _lambda_1_plus_7_tau * ( surface[i-1][j]  + surface[i][j+1] ) +
			    _2_lambda_tau * ( surface[i-1][j-1] + surface[i-1][j+1] + surface[i+1][j-1] + surface[i+1][j+1] ) +
			    _lambda_tau * ( surface[i-2][j] + surface[i][j+2] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
	    }
	  else
	  if((i==1)&&(j==Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_7_tau * ( surface[i][j-1] + surface[i+1][j] ) -
			    _lambda_1_plus_5_tau * ( surface[i-1][j] + surface[i][j+1] ) + 
			    _2_lambda_tau * ( surface[i-1][j-1] + surface[i-1][j+1] + surface[i+1][j-1] + surface[i+1][j+1] ) +
			    _lambda_tau * ( surface[i+2][j] + surface[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
            }
	  else
	  if((i==Row-2)&&(j==Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_14_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_7_tau * ( surface[i][j-1] + surface[i-1][j] ) -
			    _lambda_1_plus_5_tau * ( surface[i+1][j] + surface[i][j+1] ) +
			    _2_lambda_tau * ( surface[i-1][j-1] + surface[i-1][j+1] + surface[i+1][j-1] + surface[i+1][j+1] ) +
			    _lambda_tau * ( surface[i-2][j] + surface[i][j-2] ) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
	    }
	  else
	  if((i==1)&&(j>1)&&(j<Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_7_tau * ( surface[i][j-1] + surface[i+1][j] + surface[i][j+1] ) -
			    _lambda_1_plus_5_tau * surface[i-1][j]  +
			    _2_lambda_tau * ( surface[i-1][j-1] + surface[i-1][j+1] + surface[i+1][j-1] + surface[i+1][j+1] ) +
			    _lambda_tau * ( surface[i][j+2] + surface[i][j-2] + surface[i+2][j]) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
	    }
	  else
	  if((i==Row-2)&&(j>1)&&(j<Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_7_tau * ( surface[i][j-1] + surface[i-1][j] + surface[i][j+1] ) -
			    _lambda_1_plus_5_tau * surface[i+1][j] +
			    _2_lambda_tau * ( surface[i-1][j-1] + surface[i-1][j+1] + surface[i+1][j-1] + surface[i+1][j+1] ) +
		            _lambda_tau * ( surface[i][j-2] + surface[i][j+2] + surface[i-2][j]) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
	    }
	  else
	  if((j==1)&&(i>1)&&(i<Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_5_tau * surface[i][j-1] -
			    _lambda_1_plus_7_tau *(surface[i-1][j] + surface[i][j+1] +surface[i+1][j] ) +
			    _2_lambda_tau * ( surface[i-1][j-1] + surface[i-1][j+1] + surface[i+1][j-1] + surface[i+1][j+1] ) +
			    _lambda_tau * ( surface[i][j+2] + surface[i-2][j] + surface[i+2][j]) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
	    }
	  else
	  if((j==Column-2)&&(i>1)&&(i<Column-2))
	    {
	     T=(float)beta[i][j] + _lambda_4_plus_15_tau ;
	     surface[i][j]=surface[i][j] - (w/T) *
			   ( T * surface[i][j] -
			    _lambda_1_plus_7_tau *
			    ( surface[i][j-1] + surface[i-1][j] + surface[i+1][j] ) -
			    _lambda_1_plus_5_tau * surface[i][j+1] +
			     _2_lambda_tau *
			    ( surface[i-1][j-1] + surface[i-1][j+1] + surface[i+1][j-1] + surface[i+1][j+1] ) +
			    _lambda_tau *( surface[i][j-2] + surface[i-2][j] + surface[i+2][j]) -
			    ((float) beta[i][j]) * ((float) depth[i][j]) );
		    }
	  if(surface[i][j]>255.0) surface[i][j]=255.0 ;
	  if(surface[i][j]<0.0) surface[i][j]=0.0     ;
	  abs_error=fabs(surface[i][j]-uc) ;
	  error+=abs_error;
	 } /* for */
	 error= error / ( (float) Row * (float) Column ) ;
	 IterationNumber++;
	}while(error>=d_error);

       if(IsMemoryAllocatedForY == false){
         IsMemoryAllocatedForY = true   ;
         Y = new unsigned char*[Row] ;
         Y[0] = new unsigned char[Row*Column] ;
         unsigned char *p= Y[0]+Column;
         for (i=1;i<Row;i++,p+=Column)
             Y[i]= p;
       }
       pSurface=surface[0];
       pBeta= beta[0];
       pDepth= Y[0];
       for (i=0;i<Row;i++){
	   for (j=0;j<Column;j++,pBeta++,pSurface++,pDepth++){
     	       if ((*pBeta==0)&&(boolInterpolate==true)){
                  if (rand()%2)
	             *pSurface += 0.5;
                  else
                     *pSurface -= 0.5;
	          if (*pSurface>255.0) *pSurface= 255.0;
		  if (*pSurface<0.0) *pSurface= 0.0;
               }
               *pDepth= (unsigned char) *pSurface;
	   }
       }
}

void SOR::AllocateMemoryForModelMatrices(void)
{
     int i,j;
     unsigned char *p;
     beta = new unsigned char*[Row];
     beta[0] = new unsigned char[Row*Column];
     p=beta[0]+Column;
     for (i=1;i<Row;i++,p+=Column)
         beta[i]= p;

     depth = new unsigned char*[Row];
     depth[0] = new unsigned char[Row*Column];
     p=depth[0]+Column;
     for (i=1;i<Row;i++,p+=Column)
         depth[i]= p;
     memset(depth[0],0,sizeof(unsigned char)*Row*Column);
     memset(beta[0],0,sizeof(unsigned char)*Row*Column);
}

void SOR::SaveDepthAndBetaToTImage(TImage * img1,TImage * img2){
     int i,j    ;

     img1->Hint   = "Depth Data"                                    ;
     img2->Hint   = "Sparseness Data"                               ;
     img1->Width  = Width                                           ;
     img2->Width  = Width                                           ;
     img1->Height = Height                                          ;
     img2->Height = Height                                          ;
     img2->Left   = Width + 20                                      ;
     unsigned char *p,*q,*r,*b,cl;
     p= depth[0] ;
     b= beta[0] ;
     for (i=0;i<Row;i++){
         q= (unsigned char*) img1->Picture->Bitmap->ScanLine[i];
         //r= (unsigned char*) img2->Picture->Bitmap->ScanLine[i];
         for (j=0;j<Column;j++,p++,b++){
             cl= *p * 255;
             *q++= cl ;
             *q++= cl ;
             *q++= cl ;
             /*
             if (*b){
                *r++=0;
                *r++=0;
                *r++=0;
             }
             else
               r+=3;
               */
         }
     }
     img1->Refresh();
     img2->Refresh();        
}
